program chart;

{
************************************************************

CHART - Electronic Chart System


(c) Heiko Purnhagen 1993-1997
e-mail: purnhage@tnt.uni-hannover.de
WWW:    http://www.fet.uni-hannover.de/purnhage/


All usual disclaimers apply! You use CHART at your own risk!

This program may be copied according to the GNU General Public Licence.

CHART is written in TP6. Please check README for further details.


Changes:

93xxxx   born
950805   started change log
960719   added NMEA GPRMC
960724   added bitmap
960726   added estimated error ring
970606   GNU GPL version


************************************************************
}


{ use graftabl.com to init special graph characters }


{ due to the chart tpu units included, use Options->Compiler and then }
{ Compile->Build instead of n+ e- or n- e- }

{ important options: b- }
{ suggested options: r- d- l- }


uses
  graph,crt,dos,
  mousedrv,chargraf,chargps,chartran,grafbmp;

const
  title = 'Electronic Chart System';
  version = 'CHART 1.0';
  author = '(c) Heiko Purnhagen 1993-1997';

  filecomment = '. chart';

  initovl = 'theworld.cha';

  initminmemleft = 50*1024;

  ms = 1;   { 1/6 mark size }

  lllength = 30;

  clipfact = 10;
  maxfind = 10;   { max dist for findcoord }
  maxgridlines = 50;
  defscalelimitfact = 10;

  notdisplayed = 32000.0;

  mousedel = 20;

type
  string80 = string[80];

  coordtypet = (cline,clineend,cpoly,cpolyend,ctext,cmark);
  filetypet = (fcha,fovl,flog,fmem,fwpt);

  coordptr = ^coordt;
  blockptr = ^blockt;
  coordt = record
             ll : llt;
             xy : xyt;
             uv : uvt;
             name : blockptr;
             coordtype : coordtypet;
             pred,succ : coordptr;
           end;
  blockt = record
             name : string80;
             scalelimit : real;
             typec : char;
             coord : coordptr;
             filetype : filetypet;
             pred,succ : blockptr;
           end;

  editmodet = (maddchaline,maddchapoly,maddchatext,maddchamark,
               maddlog,maddmem,maddwpt,
               mmove,medit,minsert,mremove,mdelete);


var
  graphdriver,graphmode : integer;

  loadinitovl : boolean;

  chartcenter : llt;
  chartwidth : real;
  griddist : real;
  llar : llart;
  lloff : llt;
  currentchartsize : real;
  currentchartcenter : llt;
  defscalelimit : real;

  gpsll : llt;
  gpscog,gpssog,gpsdist,gpsbrg : real;
  gpsestimerr : real;
  gpsmsg : string;
  gpsheadtime : real;
  showgps,gpsmark,gpsmarktoggle,gpsmarkcircle : boolean;

  gpson : boolean;
  oldusemouse : boolean;
  oldsersaved : boolean;
  gpscom : integer;
  gpsnmea : boolean;
  gpslogtime : integer;
  gpslogon : boolean;
  gpsloglastcoord : coordptr;
  gpsnextlogtime : real;
  gpsbytesleft : integer;
  gpsmsgchar : string[6];
  gpsmsgstr : string;
  gpsmsglist : array[1..10] of integer;
  gpsmsgidx,gpsmsgnum : integer;

  llformat : integer;
  displayvector : boolean;
  displaytext : boolean;
  displayovl : boolean;
  displayscalelimit : boolean;
  coordfirst,coordlast : coordptr;
  blockfirst,blocklast : blockptr;
  actcoord : coordptr;
  actblock : blockptr;

  minmemleft : longint;
  initialmemavail : longint;

  fllformat,flllength : integer;

  filename,filemask : array [filetypet] of string;
  filemod : array [filetypet] of boolean;

  top1x,top2x,top3x,top4x : integer;
  bot1x,bot2x,bot3x,bot4x,bot5x,bot6x : integer;

  menufile,menudisp,menunav,menuedit : menut;
  menumenu : menut;
  menufload,menufsave,menufclear,menufbmp : menut;
  menudscale,menudconfig : menut;
  menueadd,menueedit : menut;

  ll : llt;
  xy : xyt;
  uv : uvt;

  editmode : editmodet;
  editmodename : array [editmodet] of string20;

  coordtypename : array [coordtypet] of string20;
  filetypename : array [filetypet] of string20;

  actmenu : char;
  doquit : boolean;

  bmpdisp,bmpdisplast : boolean;
  bmpbuf : array [0..3] of pointer;
  bmpreverse : boolean;
  bmpfact : real;
  bmpxoff,bmpyoff,bmpx0,bmpy0 : integer;
  bmpnamesave : string;
  bmpnameload,bmcnameload : string;
  bmpthr,bmpzoom,bmpcol : integer;
  bmpxcenter,bmpycenter : integer;
  bmpllar : llart;
  bmpuvar : uvart;
  bmpcenterll : llt;


{ procedures and functions }

{ menu handling procedures }

function b2c (b : boolean) : char;
begin
  if b then
    b2c := 'y'
  else
    b2c := 'n';
end;

function c2b (c : char; var b : boolean) : boolean;
begin
  c2b := true;
  if c = 'y' then
    b := true
  else if c = 'n' then
    b := false
  else
    c2b := false;
end;

procedure initmenus;
  var
    xx,yy : integer;
begin
  top1x := 0;
  ll.lat := 0;
  ll.lon := 0;
  top2x := textwidth(ll2str(ll))+tx1+tx2+6;
  menumaxsize := 0;
  initm (top2x,0,false);
  nextm ('m','Menu');
  nextm ('l','Load');
  nextm ('s','Save');
  nextm ('c','Clear');
  nextm ('i','Info');
  nextm ('f','Format');
  nextm ('b','Bitmap');
  quitm (menufile);
  initm (top2x,0,false);
  nextm ('m','Menu');
  nextm ('s','Scale');
  nextm ('i','zoomIn');
  nextm ('o','zoomOut');
  nextm ('z','Zoom');
  nextm ('c','Config');
  quitm (menudisp);
  initm (top2x,0,false);
  nextm ('m','Menu');
  nextm ('g','Gps');
  nextm ('i','Info');
  nextm ('w','Wpts');
  nextm ('l','Log');
  nextm ('c','Center');
  nextm ('d','Display');
  quitm (menunav);
  initm (top2x,0,false);
  nextm ('m','Menu');
  nextm ('a','Add');
  nextm ('e','Edit');
  nextm ('v','moVe');
  quitm (menuedit);

  menumaxsize := 0;
  yy := menufile.y1+menufile.eheight+6;
  xx := menufile.x1+txb;
  initm (xx,yy,true);
  nextm ('f','File');
  nextm ('d','Display');
  nextm ('n','Navigate');
  nextm ('e','Edit');
  nextm ('h','Help');
  nextm ('r','Refresh');
  nextm ('q','Quit');
  quitm (menumenu);

  xx := xx+menufile.ewidth[1];
  initm (xx,yy,true);
  nextm ('c','Chart');
  nextm ('o','Overlay');
  nextm ('l','Log');
  nextm ('m','Memo');
  nextm ('w','Waypoint');
  quitm (menufload);
  xx := xx+menufile.ewidth[2];
  initm (xx,yy,true);
  nextm ('c','Chart');
  nextm ('l','Log');
  nextm ('m','Memo');
  nextm ('w','Waypoint');
  quitm (menufsave);
  xx := xx+menufile.ewidth[3];
  initm (xx,yy,true);
  nextm ('c','Chart');
  nextm ('o','Overlay');
  nextm ('l','Log');
  nextm ('m','Memo');
  nextm ('w','Waypoint');
  nextm ('a','All');
  quitm (menufclear);
  xx := xx+menufile.ewidth[4];
  xx := xx+menufile.ewidth[5];
  xx := xx+menufile.ewidth[6];
  initm (xx,yy,true);
  nextm ('a','Align');
  nextm ('l','Load');
  nextm ('r','Reload');
  nextm ('c','Clear');
  nextm ('s','Save screen');
  quitm (menufbmp);

  xx := menudisp.x1+txb+menudisp.ewidth[1];
  initm (xx,yy,true);
  nextm ('f','Fixpoints');
  nextm ('a','Align');
  nextm ('c','Center & size');
  nextm ('m','north up Mercator');
  nextm ('l','Last');
  quitm (menudscale);
  xx := xx+menudisp.ewidth[2]+menudisp.ewidth[3]+
           menudisp.ewidth[4]+menudisp.ewidth[5];
  initm (xx,yy,true);
  nextm ('f','Format lat&lon');
  nextm ('d','gps Display');
  nextm ('g','Grid');
  nextm ('i','displayed Items');
  nextm ('r','Reverse display');
  nextm ('m','Mouse & cursor');
  quitm (menudconfig);

  xx := menuedit.x1+txb+menuedit.ewidth[1];
  initm (xx,yy,true);
  nextm ('l','chart Line');
  nextm ('p','chart Poly');
  nextm ('t','chart Text');
  nextm ('m','chart Mark');
  nextm ('o','lOg line');
  nextm ('e','mEmo mark');
  nextm ('w','Wpt mark');
  quitm (menueadd);
  xx := xx+menuedit.ewidth[2];
  initm (xx,yy,true);
  nextm ('e','Edit');
  nextm ('i','Insert point');
  nextm ('r','Remove point');
  nextm ('d','Delete');
  quitm (menueedit);

  getmem (menumenu.background,menumaxsize);
  menufload.background := menumenu.background;
  menufsave.background := menumenu.background;
  menufclear.background := menumenu.background;
  menufbmp.background := menumenu.background;
  menudscale.background := menumenu.background;
  menudconfig.background := menumenu.background;
  menueadd.background := menumenu.background;
  menueedit.background := menumenu.background;
  top3x := menufile.x2+1;
  top4x := xmax+1;
end;


{ coordinate transform }

function calcang (x,y : real) : real;
{ angle in ]-pi,pi] }
begin
  if x = 0 then
  begin
    if y = 0 then
      calcang := 0
    else
      if y > 0 then
        calcang := pi/2
      else
        calcang := -pi/2;
  end
  else
    if x > 0 then
      calcang := arctan(y/x)
    else
      if y >= 0 then
        calcang := arctan(y/x)+pi
      else
        calcang := arctan(y/x)-pi;
end;

procedure calcdistbear (ll1,ll2 : llt; var dist,brg : real);
{ rhumb line }
{ distance not precise if big !!! }
  var
    xy1,xy2 : xyt;
    lat : real;
begin
  ll2xy (ll1,xy1);
  ll2xy (ll2,xy2);
  lat := (ll1.lat+ll2.lat)/2;
  dist := sqrt(sqr(xy2.x-xy1.x)+sqr(xy2.y-xy1.y))*cos(lat);
  if dist = 0 then
    brg := 0
  else
  begin
    brg := calcang (xy2.y-xy1.y,xy2.x-xy1.x);
    if brg < 0 then
      brg := brg+2*pi;
  end;
end;

function choosegrid (dist : real) : real;
begin
  if dist <= 1 then
    choosegrid := 1
  else if dist <= 10 then
    choosegrid := 10
  else if dist <= 60 then
    choosegrid := 60
  else
    choosegrid := 600;
end;

procedure setchartsize;
  var
    diag1,diag2,b : real;
    uv1,uv2 : uvt;
    xy1,xy2 : xyt;
    ll1,ll2 : llt;
begin
  uv1.u := (xmax+1)*0.5;
  uv1.v := (ymax+1)*0.5;
  uv2xy (uv1,xy1);
  xy2ll (xy1,currentchartcenter);
  uv1.u := (xmax+1)*0.45;
  uv1.v := (ymax+1)*0.45;
  uv2xy (uv1,xy1);
  xy2ll (xy1,ll1);
  uv2.u := (xmax+1)*0.55;
  uv2.v := (ymax+1)*0.55;
  uv2xy (uv2,xy2);
  xy2ll (xy2,ll2);
  calcdistbear (ll1,ll2,diag1,b);
  uv1.u := (xmax+1)*0.55;
  uv1.v := (ymax+1)*0.45;
  uv2xy (uv1,xy1);
  xy2ll (xy1,ll1);
  uv2.u := (xmax+1)*0.45;
  uv2.v := (ymax+1)*0.55;
  uv2xy (uv2,xy2);
  xy2ll (xy2,ll2);
  calcdistbear (ll1,ll2,diag2,b);
  if diag1 > diag2 then
    currentchartsize := diag1
  else
    currentchartsize := diag2;
  currentchartsize := currentchartsize * 10 /
                      sqrt(1+sqr((ymax+1)/(xmax+1)*(yasp/xasp)));
  defscalelimit := currentchartsize*defscalelimitfact;
  if defscalelimit > 20000 then
    defscalelimit := 0;
  if (griddist <> 0) and (griddist < currentchartsize/maxgridlines) then
    griddist := choosegrid(currentchartsize/maxgridlines);
end;

procedure chartfindab (chartcenter : llt; chartwidth : real);
  var
    llborder : llt;
    xyar : xyart;
    uvar : uvart;
    b : boolean;
begin
  if cos(chartcenter.lat) = 0 then
  begin
    errorbeep;
    b := confirm ('Error: can''t init transformation',false);
    exit;
  end;
  ll2xy (chartcenter,xyar[1]);
  llborder := chartcenter;
  llborder.lon := llborder.lon + chartwidth/60/2/180*pi/cos(chartcenter.lat);
  ll2xy (llborder,xyar[2]);
  xyar[3] := xyar[1];
  xyar[3].y := xyar[3].y - (xyar[2].x-xyar[1].x)*
                           ((ymax+1)/(xmax+1))*(yasp/xasp);
  uvar[1].u := (xmax+1)/2;
  uvar[1].v := (ymax+1)/2;
  uvar[2].u := xmax+1;
  uvar[2].v := (ymax+1)/2;
  uvar[3].u := (xmax+1)/2;
  uvar[3].v := ymax+1;
  if not findab (xyar,uvar) then
  begin
    errorbeep;
    b := confirm ('Error: can''t init transformation',false);
  end
  else
  begin
    bmpdisp := false;
    bmpdisplast := false;
    saveab;
    setchartsize;
    recalc := true;
  end;
end;


{ bitmap }

procedure initbitmap (minleft : longint);
  var
    memneed : longint;
    i : integer;
begin
  memneed := imagesize(0,0,xmax,((ymax+1) div 4)-1);
  if 4*memneed+minleft < maxavail then
    for i := 0 to 3 do
      getmem (bmpbuf[i],memneed)
  else
    bmpbuf[0] := nil;
end;

procedure storebitmap;
  var
    i : integer;
    b : boolean;
begin
  if bmpbuf[0] <> nil then
  begin
    xhidecursor;
    for i := 0 to 3 do
      getimage (0,i*((ymax+1) div 4),xmax,(i+1)*((ymax+1) div 4)-1,
                bmpbuf[i]^);
    xshowcursor;
    bmpreverse := reverse;
  end
end;

procedure restorebitmap;
  var
    b : boolean;
    i,m : integer;
begin
  if bmpbuf[0] = nil then
    exit;
  if reverse xor bmpreverse then
    m := notput
  else
    m := normalput;
  xhidecursor;
  for i := 0 to 3 do
    putimage (0,i*((ymax+1) div 4),bmpbuf[i]^,m);
  xshowcursor;
end;

procedure calcbitmaps2b (xs,ys : real; var xb,yb : real);
begin
  xb := bmpxoff+(xs-bmpx0)/bmpfact;
  yb := bmpyoff-(ys-bmpy0)/bmpfact;
end;

procedure calcbitmapb2s (xb,yb : real; var xs,ys : real);
begin
  xs := (xb-bmpxoff)*bmpfact+bmpx0;
  ys := -(yb-bmpyoff)*bmpfact+bmpy0;
end;

function loadbitmap (fn : string; xcenter,ycenter,xsize,ysize : integer;
                     zoom,thr,bmpcol : integer; doload : boolean) : boolean;
{ OK->true error->false }
{ zoom=-999 -> fit into size}
  var
    s : string;
    wid,hig,col,t : integer;
    c00 : boolean;
    b : boolean;
    fgc,bgc : integer;
begin
  loadbitmap := false;
  s := readbmphdr (fn,wid,hig,col,c00);
  if s <> '' then
  begin
    b := confirm ('Error: '+fn+' '+s,false);
    exit;
  end;
  if zoom = -999 then
  begin
    zoom := 0;
    while wid shr -zoom > xsize do
      dec (zoom);
    while hig shr -zoom > ysize do
      dec (zoom);
  end;
  bmpfact := (1 shl (8+zoom))/256;
  t := ((wid shl 2) shr 2-zoom);
  if t < xsize then
    xsize := t;
  t := ((hig shl 2) shr 2-zoom);
  if t < ysize then
    ysize := t;
  bmpxoff := xcenter-round(xsize/2/bmpfact);
  bmpyoff := ycenter-round(ysize/2/bmpfact);
  bmpx0 := (xmax+1-xsize) div 2;
  bmpy0 := (ymax+1+ysize) div 2 - 1;
  if doload then
  begin
    newscreen;
    if not c00 then
      thr := -thr;
    if not reverse then
    begin
      bgc := 0;
      fgc := 0+bmpcol;
    end
    else
    begin
      bgc := getmaxcolor;
      fgc := bgc-bmpcol;
    end;
    xhidecursor;
    s := readbmp (fn,bmpxoff,bmpyoff,bmpx0,bmpy0,
                  xsize,ysize,zoom,thr,bgc,fgc);
    xshowcursor;
    if s <> '' then
    begin
      b := confirm ('Error: '+fn+' '+s,false);
      exit;
    end;
  end;
  loadbitmap := true;
end;

function savebitmap (fn : string) : boolean;
{ OK->true error->false }
  var
    s : string;
    b : boolean;
begin
  savebitmap := false;
  xhidecursor;
  s := writebmp (fn,0,ymax,xmax+1,ymax+1);
  xshowcursor;
  if s <> '' then
  begin
    b := confirm ('Error: '+fn+' '+s,false);
    exit;
  end;
  savebitmap := true;
end;

function loadbmc (bmcfn : string; var bmpfn : string; var thr : integer;
                  var llar : llart; var uvar : uvart) : boolean;
{ OK->true error->false }
  var
    f : text;
    s,s1,s2 : string;
    b : boolean;
    i,p : integer;
    er : integer;

  function readline (var s : string) : boolean;
  begin
    readline := false;
    {$i-}
    readln (f,s);
    {$i+}
    if ioresult <> 0 then
    begin
      b := confirm ('Error: can''t read '+bmcfn,false);
      close (f);
      exit;
    end;
    readline := true;
  end;

begin
  loadbmc := false;
  assign (f,bmcfn);
  {$i-}
  reset (f);
  {$i+}
  if ioresult <> 0 then
  begin
    b := confirm ('Error: can''t open '+bmcfn,false);
    exit;
  end;
  if not readline (s) then
    exit;
  if s <> 'bitmap chart' then
  begin
    b := confirm ('Error: no bitmap chart header',false);
    close (f);
    exit;
  end;
  if not readline (bmpfn) then
    exit;
  if s = '' then
  begin
    b := confirm ('Error: no bitmap filename',false);
    close (f);
    exit;
  end;
  if not readline (s) then
    exit;
  val (s,thr,er);
  if er <> 0 then
  begin
    b := confirm ('Error: wrong threshold format',false);
    close (f);
    exit;
  end;
  for i := 1 to 3 do
  begin
    if not readline (s) then
      exit;
    if not str2ll (s,llar[i]) then
    begin
      b := confirm ('Error: wrong coordinate format',false);
      close (f);
      exit;
    end;
  end;
  for i := 1 to 3 do
  begin
    if not readline (s) then
      exit;
    while (length(s) > 0) and (s[1] = ' ') do
      s := copy(s,2,length(s)-1);
    p := pos(' ',s);
    if p = 0 then
      er := 1
    else
    begin
      s1 := copy(s,1,p-1);
      s2 := copy(s,p+1,length(s)-p);
    end;
    if er = 0 then
      val(s1,uvar[i].u,er);
    if er = 0 then
      val(s2,uvar[i].v,er);
    if er <> 0 then
    begin
      b := confirm ('Error: wrong bmp coordinate format',false);
      close (f);
      exit;
    end;
  end;
  close (f);
  loadbmc := true;
end;

function savebmc (bmcfn : string; var bmpfn : string; var thr : integer;
                  var llar : llart; var uvar : uvart) : boolean;
{ OK->true error->false }
  var
    f : text;
    s,s1,s2 : string;
    b : boolean;
    i : integer;

  function writeline (s : string) : boolean;
  begin
    writeline := false;
    {$i-}
    writeln (f,s);
    {$i+}
    if ioresult <> 0 then
    begin
      b := confirm ('Error: can''t write '+bmcfn,false);
      close (f);
      exit;
    end;
    writeline := true;
  end;

begin
  savebmc := false;
  assign (f,bmcfn);
  {$i-}
  rewrite (f);
  {$i+}
  if ioresult <> 0 then
  begin
    b := confirm ('Error: can''t open '+bmcfn,false);
    exit;
  end;
  if not writeline ('bitmap chart') then
    exit;
  if not writeline (bmpfn) then
    exit;
  str (thr,s);
  if not writeline (s) then
    exit;
  for i := 1 to 3 do
    if not writeline (ll2str(llar[i])) then
      exit;
  for i := 1 to 3 do
  begin
    str (uvar[i].u:8:1,s1);
    str (uvar[i].v:8:1,s2);
    if not writeline (s1+' '+s2) then
      exit;
  end;
  close (f);
  savebmc := true;
end;


{ initialisations }

function initdev : boolean;
{ OK -> true, ERROR -> false}
  var
    initgraphdriver, initgraphmode : integer;
    bgipath : string;
    i,er : integer;
begin
  initdev := false;
  initgraphdriver := detect;
  reverse := false;
  bgipath := '\tp6\bgi';
  loadinitovl := true;
  usemouse := true;
  minmemleft := initminmemleft;
  i := 1;
  while i <= paramcount do
  begin
    if paramstr(i) = 'b' then
    begin
      if paramcount < i+1 then
      begin
        writeln ('Error: bgi path missing');
        exit;
      end;
      inc (i);
      bgipath := paramstr(i);
    end
    else if paramstr(i) = 'c' then
    begin
      initgraphdriver := cga;
      initgraphmode := cgahi;
    end
    else if paramstr(i) = 'e' then
    begin
      initgraphdriver := ega;
      initgraphmode := egahi;
    end
    else if paramstr(i) = 'v' then
    begin
      initgraphdriver := vga;
      initgraphmode := vgahi;
    end
    else if paramstr(i) = 'a' then
    begin
      initgraphdriver := att400;
      initgraphmode := att400hi;
    end
    else if paramstr(i) = 'r' then
      reverse := true
    else if paramstr(i) = 'l' then
      loadinitovl := false
    else if paramstr(i) = 'm' then
      usemouse := false
    else if paramstr(i) = 'd' then
    begin
      if paramcount < i+1 then
      begin
        writeln ('Error: data size missing');
        exit;
      end;
      inc (i);
      val (paramstr(i),minmemleft,er);
      if er <> 0 then
      begin
        writeln ('Error: can''t read data size');
        exit;
      end;
    end
    else
    begin
      writeln (title+'   '+version+'   '+author);
      writeln;
      if paramstr(i) <> 'h' then
      begin
        writeln ('Error: unkown option ',paramstr(i));
        writeln;
      end;
      writeln ('options:');
      writeln ('  b path   bgi path');
      writeln ('  c        cga (640*200)');
      writeln ('  e        ega (640*350)');
      writeln ('  v        vga (640*480)');
      writeln ('  a        att400 (640*400)');
      writeln ('  r        reverse display');
      writeln ('  l        don''t load init overlay');
      writeln ('  m        don''t use mouse');
      writeln ('  d size   min data memory size [byte]');
      writeln ('  h        help (this text)');
      writeln;
      writeln ('use graftabl.com to init special graph characters');
      exit;
    end;
    inc (i);
  end;
  graphdriver := initgraphdriver;
  graphmode := initgraphmode;
  initgraph (graphdriver,graphmode,bgipath);
  if graphresult <> 0 then
  begin
    graphdriver := initgraphdriver;
    graphmode := initgraphmode;
    initgraph (graphdriver,graphmode,'');
    if graphresult <> 0 then
    begin
      writeln ('Error: can''t initialize graph');
      exit;
    end;
  end;
  if usemouse then
    usemouse := mouseinit;
  initdev := true;
end;

procedure initcolor;
begin
  if not reverse then
  begin
    bgc := 0;
    fgc := getmaxcolor;
  end
  else
  begin
    bgc := getmaxcolor;
    fgc := 0;
  end;
  setcolor (fgc);
  setfillstyle (solidfill,bgc);
end;

procedure init;
  var
    i : integer;
begin
  initcolor;
  revcircleinit;
  xmax := getmaxx;
  ymax := getmaxy;
  getaspectratio (xasp,yasp);
  actmenu := 'f';
  gpson := false;
  oldsersaved := false;
  gpscom := 1;
  gpsnmea := true;
  gpslogtime := 10;
  gpslogon := false;
  gpsmsgchar := 'pv  ce';
  gpsmsgstr := 'c';
  cursorbig := true;
  mousefine := true;
  mouseslow := false;
  displayvector := true;
  displaytext := true;
  displayovl := true;
  displayscalelimit := true;
  llformat := 2;
  initll2str (llformat,lllength);
  fllformat := llformat;
  flllength := lllength;
  gpstime := -999;
  gpsll.lat := 0;
  gpsll.lon := 0;
  gpscog := 0;
  gpssog := 0;
  gpsdist := 0;
  gpsbrg := 0;
  gpsestimerr := 0;
  gpsmsg := 'gps off';
  gpsheadtime := 5;
  showgps := true;
  gpsmark := true;
  gpsmarktoggle := true;
  gpsmarkcircle := true;
  bmpdisp := false;
  bmpdisplast := false;
  bmpzoom := 0;
  bmpcol := (getmaxcolor+1) div 2;
  if bmpcol = 0 then
    bmpcol := 1;
  initfine;
  settextstyle (defaultfont,horizdir,0);
  settextjustify (lefttext,toptext);
  th := textheight(' ');
  tw := textwidth(' ');
  for i := 1 to 3 do
  begin
    llar[i].lat := 0;
    llar[i].lon := 0;
  end;
  lloff.lat := 0;
  lloff.lon := 0;
  chartcenter.lat := 0;
  chartcenter.lon := 0;
  chartwidth := 360*60;
  griddist := 600;
  chartfindab (chartcenter,chartwidth);
  bmpnamesave := 'noname.bmp';
  bmpnameload := '';
  bmcnameload := '';
  filename[fcha] := 'noname.cha';
  filename[fovl] := 'noname.cha';
  filename[flog] := 'noname.log';
  filename[fmem] := 'noname.mem';
  filename[fwpt] := 'noname.wpt';
  filemask[fcha] := '*.cha';
  filemask[fovl] := '*.cha';
  filemask[flog] := '*.log';
  filemask[fmem] := '*.mem';
  filemask[fwpt] := '*.wpt';
  filemod[fcha] := false;
  filemod[fovl] := false;
  filemod[flog] := false;
  filemod[fmem] := false;
  filemod[fwpt] := false;
  coordfirst := nil;
  coordlast := nil;
  blockfirst := nil;
  blocklast := nil;
  editmodename[maddchaline] := 'add chart line | +';
  editmodename[maddchapoly] := 'add chart poly | +';
  editmodename[maddchatext] := 'add chart text | +';
  editmodename[maddchamark] := 'add chart mark | +';
  editmodename[maddlog]     := 'add log line | +';
  editmodename[maddmem]     := 'add memo mark | +';
  editmodename[maddwpt]     := 'add wpt mark | +';
  editmodename[mmove]       := 'move | +';
  editmodename[medit]       := 'edit | +';
  editmodename[minsert]     := 'insert point | +';
  editmodename[mremove]     := 'remove point | +';
  editmodename[mdelete]     := 'delete | +';
  coordtypename[cline]    := 'line';
  coordtypename[clineend] := 'line';
  coordtypename[cpoly]    := 'poly';
  coordtypename[cpolyend] := 'poly';
  coordtypename[ctext]    := 'text';
  coordtypename[cmark]    := 'mark';
  filetypename[fcha] := 'chart';
  filetypename[fovl] := 'overlay';
  filetypename[flog] := 'log';
  filetypename[fmem] := 'memo';
  filetypename[fwpt] := 'waypoint';
  initmenus;
  initconfirm (minmemleft);
  initwindow (minmemleft);
  initbitmap (minmemleft);
  initialmemavail := memavail;
end;


{ draw grid and chart etc. }

procedure transcoord;
  var
    coord : coordptr;
begin
  coord := coordfirst;
  while coord <> nil do
  begin
    if ((coord^.name^.filetype = fovl) and not displayovl) or
       ((coord^.name^.scalelimit < currentchartsize) and
        (coord^.name^.scalelimit <> 0) and displayscalelimit) then
    begin
      coord^.uv.u := notdisplayed;
      coord^.uv.v := notdisplayed;
    end
    else
    begin
      ll2xy (coord^.ll,coord^.xy);
      xy2uv (coord^.xy,coord^.uv);
    end;
    coord := coord^.succ;
  end;
end;

function round2xy (uv : uvt; var x,y : integer) : boolean;
begin
  if (uv.u < (1+clipfact)*xmax) and (uv.u > -clipfact*xmax) and
     (uv.v < (1+clipfact)*ymax) and (uv.v > -clipfact*ymax) then
  begin
    x := round(uv.u);
    y := round(uv.v);
    round2xy := true;
  end
  else
  begin
    x := 0;
    y := 0;
    round2xy := false;
  end;
end;

procedure drawgrid;
  var
    minlat,maxlat,minlon,maxlon : real;
    uv1,uv2 : uvt;
    xy1,xy2 : xyt;
    ll1,ll2 : llt;
    raddist : real;
    x1,y1,x2,y2 : integer;

  procedure setminmax;
  begin
    uv2xy (uv1,xy1);
    xy2ll (xy1,ll1);
    if ll1.lat < minlat then
      minlat := ll1.lat;
    if ll1.lat > maxlat then
      maxlat := ll1.lat;
    if ll1.lon < minlon then
      minlon := ll1.lon;
    if ll1.lon > maxlon then
      maxlon := ll1.lon;
  end;

begin
  if griddist = 0 then
    exit;
  uv1.u := 0;
  uv1.v := 0;
  uv2xy (uv1,xy1);
  xy2ll (xy1,ll1);
  minlat := ll1.lat;
  maxlat := ll1.lat;
  minlon := ll1.lon;
  maxlon := ll1.lon;
  uv1.u := xmax;
  setminmax;
  uv1.v := ymax;
  setminmax;
  uv1.u := 0;
  setminmax;
  xhidecursor;
  raddist := griddist/60/180*pi;
  ll1.lat := minlat;
  ll2.lat := maxlat;
  ll1.lon := trunc(minlon/raddist)*raddist;
  while ll1.lon < maxlon do
  begin
    ll2.lon := ll1.lon;
    ll2xy (ll1,xy1);
    xy2uv (xy1,uv1);
    ll2xy (ll2,xy2);
    xy2uv (xy2,uv2);
    if round2xy(uv1,x1,y1) and round2xy(uv2,x2,y2) then
      line (x1,y1,x2,y2);
    ll1.lon := ll1.lon + raddist;
  end;
  ll1.lon := minlon;
  ll2.lon := maxlon;
  ll1.lat := trunc(minlat/raddist)*raddist;
  while ll1.lat < maxlat do
  begin
    ll2.lat := ll1.lat;
    ll2xy (ll1,xy1);
    xy2uv (xy1,uv1);
    ll2xy (ll2,xy2);
    xy2uv (xy2,uv2);
    if round2xy(uv1,x1,y1) and round2xy(uv2,x2,y2) then
      line (x1,y1,x2,y2);
    ll1.lat := ll1.lat + raddist;
  end;
  xshowcursor;
end;

procedure drawlines;
  var
    coord : coordptr;
    x,y,xp,yp : integer;
    drawing,ispoly,firstpoly,closepoly : boolean;
    c : char;
begin
  xhidecursor;
  drawing := false;
  ispoly := false;
  coord := coordfirst;
  while coord <> nil do
  begin
    if displayovl or (coord^.name^.filetype <> fovl) then
    begin
      if coord^.coordtype in [cline,clineend,cpoly,cpolyend] then
      begin
        c := coord^.name^.typec;
        if (c < 'a') or (c > 'e') then
          c := 'a';
        case c of
          'a' : setlinestyle (solidln,0,normwidth);
          'b' : setlinestyle (dottedln,0,normwidth);
          'c' : setlinestyle (centerln,0,normwidth);
          'd' : setlinestyle (dashedln,0,normwidth);
          'e' : setlinestyle (solidln,0,thickwidth);
        end;
        if coord^.coordtype in [cpoly,cpolyend] then
        begin
          firstpoly := not ispoly;
          if firstpoly then
            closepoly := false;
          ispoly := true;
        end;
        if round2xy (coord^.uv,x,y) then
        begin
          if firstpoly then
          begin
            xp := x;
            yp := y;
            closepoly := true;
          end;
          if drawing then
            lineto (x,y)
          else
            moveto (x,y);
          linerel (0,0);
          drawing := true;
        end
        else
          drawing := false;
      end;
      if (coord^.coordtype = cpolyend) and closepoly and drawing then
        lineto (xp,yp);
      if not (coord^.coordtype in [cline,cpoly]) then
      begin
        drawing := false;
        ispoly := false;
      end;
    end;
    coord := coord^.succ;
  end;
  setlinestyle (solidln,0,normwidth);
  xshowcursor;
end;

procedure drawmark (x,y : integer; filetype : filetypet);
begin
  moveto (x,y);
  case filetype of
    fcha : begin
             moverel (3*ms,0);
             linerel (-6*ms,0);
             moverel (3*ms,3*ms);
             linerel (0,-6*ms);
           end;
    fovl : begin
             moverel (3*ms,0);
             linerel (-6*ms,0);
             moverel (3*ms,3*ms);
             linerel (0,-6*ms);
           end;
    fmem : begin
             moverel (2*ms,2*ms);
             linerel (-4*ms,-4*ms);
             moverel (0,4*ms);
             linerel (4*ms,-4*ms);
           end;
    fwpt : begin
             moverel (3*ms,-1*ms);
             linerel (0,2*ms);
             linerel (-2*ms,2*ms);
             linerel (-2*ms,0);
             linerel (-2*ms,-2*ms);
             linerel (0,-2*ms);
             linerel (2*ms,-2*ms);
             linerel (2*ms,0);
             linerel (2*ms,2*ms);
           end;
  end;
end;

procedure drawmarks;
  var
    coord : coordptr;
    x,y : integer;
begin
  xhidecursor;
  coord := coordfirst;
  while coord <> nil do
  begin
    if displayovl or (coord^.name^.filetype <> fovl) then
      if coord^.coordtype = cmark then
        if round2xy (coord^.uv,x,y) then
          drawmark (x,y,coord^.name^.filetype);
    coord := coord^.succ;
  end;
  xshowcursor;
end;

procedure getdrawtext (coordtype : coordtypet; s : string; c : char;
                       var x,y : integer);
begin
  if (c < 'a') or (c > 'i') then
  begin
    if coordtype = ctext then
      c := 'e'
    else
      c := 'i';
  end;
  if c in ['a','d','g'] then
    x := x-3*ms-2-textwidth(s);
  if c in ['b','e','h'] then
    x := x+1-textwidth(s) div 2;
  if c in ['c','f','i'] then
    x := x+3*ms+4;
  if c in ['a','b','c'] then
    y := y-3*ms-2-textheight(s);
  if c in ['d','e','f'] then
    y := y+1-textheight(s) div 2;
  if c in ['g','h','i'] then
    y := y+3*ms+4;
end;

procedure drawtexts;
  var
    coord : coordptr;
    x,y : integer;
    s : string;
begin
  xhidecursor;
  coord := coordfirst;
  while coord <> nil do
  begin
    if displayovl or (coord^.name^.filetype <> fovl) then
      if coord^.coordtype in [ctext,cmark] then
        if round2xy (coord^.uv,x,y) then
        begin
          s := coord^.name^.name;
          getdrawtext (coord^.coordtype,s,coord^.name^.typec,x,y);
          moveto (x,y);
          outtext (s);
        end;
    coord := coord^.succ;
  end;
  xshowcursor;
end;


{ clear, load & save of charts etc. }

procedure removecoord (coord : coordptr);
begin
  if coord^.pred = nil then
    coordfirst := coord^.succ
  else
    coord^.pred^.succ := coord^.succ;
  if coord^.succ = nil then
    coordlast := coord^.pred
  else
    coord^.succ^.pred := coord^.pred;
end;

procedure removeblock (block : blockptr);
begin
  if block^.pred = nil then
    blockfirst := block^.succ
  else
    block^.pred^.succ := block^.succ;
  if block^.succ = nil then
    blocklast := block^.pred
  else
    block^.succ^.pred := block^.pred;
end;

procedure insertcoord (coord,after : coordptr);
{ after = nil -> insert at the begin ... }
begin
  coord^.pred := after;
  if after = nil then
  begin
    coord^.succ := coordfirst;
    if coordfirst = nil then
      coordlast := coord
    else
      coordfirst^.pred := coord;
    coordfirst := coord;
  end
  else
  begin
    coord^.succ := after^.succ;
    if after^.succ = nil then
      coordlast := coord
    else
      after^.succ^.pred := coord;
    after^.succ := coord;
  end;
end;

procedure insertblock (block,after : blockptr);
{ after = nil -> insert at the begin ... }
begin
  block^.pred := after;
  if after = nil then
  begin
    block^.succ := blockfirst;
    if blockfirst = nil then
      blocklast := block
    else
      blockfirst^.pred := block;
    blockfirst := block;
  end
  else
  begin
    block^.succ := after^.succ;
    if after^.succ = nil then
      blocklast := block
    else
      after^.succ^.pred := block;
    after^.succ := block;
  end;
end;

procedure cleard (filetype : filetypet);
  var
    actblock,nextblock : blockptr;
    actcoord,nextcoord : coordptr;
begin
  actblock := blockfirst;
  while actblock <> nil do
  begin
    nextblock := actblock^.succ;
    if actblock^.filetype = filetype then
    begin
      actcoord := actblock^.coord;
      while actcoord^.coordtype in [cline,cpoly] do
      begin
        nextcoord := actcoord^.succ;
        removecoord (actcoord);
        dispose (actcoord);
        actcoord := nextcoord;
      end;
      removecoord (actcoord);
      dispose (actcoord);
      removeblock (actblock);
      dispose (actblock);
    end;
    actblock := nextblock;
  end;
  recalc := true;
end;

function loadd (filename : string; filetype : filetypet) : boolean;
  var
    f : text;
    s,si : string;
    b,found : boolean;
    linenum : integer;

  function readlnok (var s : string; expectline : boolean) : boolean;
    var
      b : boolean;
  begin
    if eof(f) then
    begin
      if expectline then
      begin
        b := confirm ('Error: unexpected end of '+filename,false);
      end;
      readlnok := false;
      exit;
    end;
    {$i-}
    readln (f,s);
    {$i+}
    if ioresult <> 0 then
    begin
      b := confirm ('Error: can''t read '+filename,false);
      readlnok := false;
    end
    else
    begin
      inc (linenum);
      readlnok := true;
    end;
  end;

  function loadblock (var s : string; coordtype : coordtypet) : boolean;
    var
      actblock : blockptr;
      actcoord : coordptr;
      scalelimit : real;
      typec : char;
      ll : llt;
      b : boolean;
      er : integer;
  begin
    loadblock := true;
    if copy(s,1,length(coordtypename[coordtype])) <>
       coordtypename[coordtype] then
      exit;
    found := true;
    scalelimit := 0;
    typec := ' ';
    delete (s,1,length(coordtypename[coordtype]));
    while (s <> '') and (s[1] in [' ',tab]) do
      delete (s,1,1);
    if (s <> '') and not (s[1] in ['0'..'9','.']) then
    begin
      typec := s[1];
      delete (s,1,1);
    end;
    while (s <> '') and (s[1] in [' ',tab]) do
      delete (s,1,1);
    while (s <> '') and (s[length(s)] in [' ',tab]) do
      delete (s,length(s),1);
    if s <> '' then
    begin
      val (s,scalelimit,er);
      if er <> 0 then
      begin
        str (linenum,si);
        b := confirm ('Error: can''t read scalelimit in line '+si+' of '+
                      filename,false);
        loadblock := false;
        exit;
      end;
    end;
    if not readlnok (s,true) then
    begin
      loadblock := false;
      exit;
    end;
    new (actblock);
    actblock^.filetype := filetype;
    actblock^.name := s;
    actblock^.scalelimit := scalelimit;
    actblock^.typec := typec;
    if not readlnok (s,true) then
    begin
      dispose (actblock);
      loadblock := false;
      exit;
    end;
    if not str2ll (s,ll) then
    begin
      str (linenum,si);
      b := confirm ('Error: wrong coordinate format in line '+si+' of '+
                    filename,false);
      dispose (actblock);
      loadblock := false;
      exit;
    end;
    new (actcoord);
    actcoord^.name := actblock;
    actblock^.coord := actcoord;
    actcoord^.coordtype := coordtype;
    actcoord^.ll := ll;
    insertblock (actblock,blocklast);
    insertcoord (actcoord,coordlast);
    if not (coordtype in [cline,cpoly]) then
      loadblock := readlnok (s,false)
    else
      while true do
      begin
        repeat
          if not readlnok (s,false) then
          begin
            if coordtype = cline then
              actcoord^.coordtype := clineend
            else
              actcoord^.coordtype := cpolyend;
            loadblock := false;
            exit;
          end;
        until (s <> '') and (copy(s,1,1) <> '.');
        if not str2ll (s,ll) then
        begin
          if coordtype = cline then
            actcoord^.coordtype := clineend
          else
            actcoord^.coordtype := cpolyend;
          loadblock := true;
          exit;
        end;
        new (actcoord);
        actcoord^.name := actblock;
        actcoord^.coordtype := coordtype;
        actcoord^.ll := ll;
        insertcoord (actcoord,coordlast);
      end;
  end;

  procedure closefile;
  begin
    {$i-}
    close (f);
    {$i+}
    if ioresult <> 0 then
    begin
      b := confirm ('Error: can''t close '+filename,false);
      exit;
    end;
  end;

begin
  loadd := false;
  assign (f,filename);
  {$i-}
  reset (f);
  {$i+}
  if ioresult <> 0 then
  begin
    b := confirm ('Error: can''t open '+filename,false);
    exit;
  end;
  recalc := true;
  loadd := true;
  s := '';
  linenum := 0;
  while true do
  begin
    while (s = '') or (copy(s,1,1) = '.') do
      if not readlnok (s,false) then
      begin
        closefile;
        exit;
      end;
    found := false;
    if filetype in [fcha,fovl,flog] then
      if not loadblock (s,cline) then
      begin
        closefile;
        exit;
      end;
    if filetype in [fcha,fovl] then
      if not loadblock (s,cpoly) then
      begin
        closefile;
        exit;
      end;
    if filetype in [fcha,fovl] then
      if not loadblock (s,ctext) then
      begin
        closefile;
        exit;
      end;
    if filetype in [fcha,fovl,fmem,fwpt] then
      if not loadblock (s,cmark) then
      begin
        closefile;
        exit;
      end;
    if not found then
    begin
      str (linenum,si);
      b := confirm ('Error: unknown block in line '+si+' of '+filename,false);
      exit;
    end;
  end;
  closefile;
end;

function saved (filename : string; filetype : filetypet) : boolean;
  var
    f : text;
    actblock : blockptr;
    actcoord : coordptr;
    s,ss : string;
    b : boolean;

  function writelnok (s : string) : boolean;
    var
      b : boolean;
  begin
    {$i-}
    writeln (f,s);
    {$i+}
    if ioresult <> 0 then
    begin
      b := confirm ('Error: can''t write '+filename,false);
      writelnok := false;
    end
    else
      writelnok := true;
  end;

begin
  saved := false;
  assign (f,filename);
  {$i-}
  reset (f);
  {$i+}
  if ioresult = 0 then
    if not confirm (filename+' exists! overwrite?',true) then
      exit;
  {$i-}
  rewrite (f);
  {$i+}
  if ioresult <> 0 then
  begin
    b := confirm ('Error: can''t open '+filename,false);
    exit;
  end;
  if not writelnok (filecomment) then
    exit;
  actblock := blockfirst;
  while actblock <> nil do
  begin
    if actblock^.filetype = filetype then
    begin
      actcoord := actblock^.coord;
      s := coordtypename[actcoord^.coordtype];
      if actblock^.typec <> ' ' then
        s := s+' '+actblock^.typec;
      if actblock^.scalelimit <> 0 then
      begin
        str (actblock^.scalelimit:6:4,ss);
        s := s+' '+ss;
      end;
      if not writelnok (s) then
        exit;
      if not writelnok (actblock^.name) then
        exit;
      while actcoord^.coordtype in [cline,cpoly] do
      begin
        if not writelnok (ll2str(actcoord^.ll)) then
          exit;
        actcoord := actcoord^.succ;
      end;
      if not writelnok (ll2str(actcoord^.ll)) then
        exit;
    end;
    actblock := actblock^.succ;
  end;
  {$i-}
  close (f);
  {$i+}
  if ioresult <> 0 then
  begin
    b := confirm ('Error: can''t close '+filename,false);
    exit;
  end;
  saved := true;
end;


{ several procedures }

procedure newactpos;
  var
    s : string;
begin
  uv.u := cx;
  uv.v := cy;
  uv2xy (uv,xy);
  xy2ll (xy,ll);
  s := ll2str (ll);
  newtext ((top1x+top2x) div 2,3,true,s);
end;

procedure initactpos;
  var
    x,y : integer;
begin
  boxtext (top1x,0,top2x-1,th+ty1+ty2+5,true,'');
  newactpos;
end;

procedure showmode (s : string);
begin
  boxtext (top3x,0,top4x-1,th+ty1+ty2+5,true,s);
end;

function brg2str (brg : real) : string;
  var
    s : string;
begin
  str ((brg/pi*180):5:1,s);
  if s[1] = ' ' then
    s[1] := '0';
  if s[2] = ' ' then
    s[2] := '0';
  if s = '360.0' then
    s := '000.0';
  brg2str := s+'';
end;

procedure newgpscur;
  var
    s,ss : string;
    y : integer;
begin
  y := ymax-th-ty1-ty2-2;
  s := brg2str(gpsbrg);
  str (gpsdist:5:3,ss);
  if length(ss) > 5 then
    str (gpsdist:5:2,ss);
  if length(ss) > 5 then
    str (gpsdist:5:1,ss);
  if length(ss) > 5 then
    str (gpsdist:5:0,ss);
  s := s+' '+ss+'nm';
  newtext ((bot3x+bot4x) div 2,y,true,s);
end;

procedure newgpsdata (onlymsg : boolean);
  var
    s,ss : string;
    y : integer;
begin
  boxtext (bot4x,ymax-th-ty1-ty2-5,bot5x-1,ymax,true,gpsmsg);
  if not onlymsg then
  begin
    y := ymax-th-ty1-ty2-2;
    s := ll2str(gpsll);
    newtext ((bot1x+bot2x) div 2,y,true,s);
    s := brg2str(gpscog);
    str (gpssog:4:1,ss);
    s := s+' '+ss+'kn';
    newtext ((bot2x+bot3x) div 2,y,true,s);
    str (currentchartsize:5:3,ss);
    if length(ss) > 5 then
      str (currentchartsize:5:2,ss);
    if length(ss) > 5 then
      str (currentchartsize:5:1,ss);
    if length(ss) > 5 then
      str (currentchartsize:5:0,ss);
    s := ss+'nm';
    newtext ((bot5x+bot6x) div 2,y,true,s);
    newgpscur;
  end;
end;

procedure initgpsdata;
  var
    y1,y2 : integer;
    s : string;
begin
  y2 := ymax;
  y1 := y2-th-ty1-ty2-5;
  s := ll2str(gpsll);
  bot1x := 0;
  bot2x := bot1x+textwidth(s)+tx1+tx2+6;
  s := '000.0 00.0kn';
  bot3x := bot2x+textwidth(s)+tx1+tx2+6;
  s := '000.0 0.000nm';
  bot4x := bot3x+textwidth(s)+tx1+tx2+6;
  s := '0.000nm';
  bot6x := xmax+1;
  bot5x := bot6x-textwidth(s)-tx1-tx2-6;
  boxtext (bot1x,y1,bot2x-1,y2,false,'');
  boxtext (bot2x,y1,bot3x-1,y2,false,'');
  boxtext (bot3x,y1,bot4x-1,y2,false,'');
  boxtext (bot4x,y1,bot5x-1,y2,false,'');
  boxtext (bot5x,y1,bot6x-1,y2,false,'');
  newgpsdata (false);
end;


{ main procedures }

procedure drawchart;
begin
  newscreen;
  if bmpdisp then
    if bmpbuf[0] <> nil then
      restorebitmap
    else
      if confirm ('reload bitmap',true) then
        if not loadbitmap (bmpnameload,bmpxcenter,bmpycenter,
                           xmax+1,ymax+1,bmpzoom,bmpthr,
                           bmpcol,true) then
        begin
          newscreen;
          bmpdisp := false;
          bmpdisplast := false;
        end;
  drawgrid;
  if not displayvector then
    exit;
  drawlines;
  drawmarks;
  if displaytext then
    drawtexts;
end;

procedure titlescreen;
  var
    s : string;
    d,m,y,dow : word;
begin
  newscreen;
  if loadinitovl then
    if loadd (initovl,fovl) then
      transcoord;
  drawchart;
  s := title+'   '+version+'   '+author;
  boxtext (0,0,xmax,th+ty1+ty2+5,true,s);
  openwindow;
  newtext (winxc,winyr[1],true,
           'Welcome to CHART - an Electronic Chart System');
  newtext (winxc,winyr[3],true,
           'All usual disclaimers apply! You use CHART at your own risk!');
  newtext (winxc,winyr[5],true,
           'CHART may be copied according to the GNU General Public Licence.');
  newtext (winxc,winyr[7],true,
           'written by Heiko Purnhagen              ');
  newtext (winxc,winyr[8],true,
           'e-mail: purnhage@tnt.uni-hannover.de    ');
  newtext (winxc,winyr[9],true,
           'http://www.fet.uni-hannover.de/purnhage/');
{
  getdate (y,m,d,dow);

  if (y > expirey) or
     ((y = expirey) and (m > expirem)) or
     ((y = expirey) and (m = expirem) and (d > expired)) then
  begin
    newtext (winxc,winyr[9],true,'>>> program expired <<<');
    waitclickkey;
    closewindow;
    doquit := true;
    exit;
  end;
  newtext (winxc,winyr[9],true,
           'program expires '+date2str(expired,expirem,expirey));
}
  delay (2000);
  waitclickkey;
  closewindow;
end;

procedure drawgpsmark;
  var
    xy : xyt;
    uv : uvt;
    dl,dx,dy : real;
    x,y,xh,yh : integer;
    b : boolean;
    errrad : integer;

begin
  if not (gpsmark and gpsmarktoggle) then
    exit;
  ll2xy (gpsll,xy);
  xy2uv (xy,uv);
  if not round2xy(uv,x,y) then
    exit;
  dl := gpssog/60*gpsheadtime/cos(gpsll.lat);
  dx := dl*sin(gpscog);
  dy := dl*cos(gpscog);
  xy.x := xy.x+dx;
  xy.y := xy.y+dy;
  xy2uv (xy,uv);
  b := round2xy(uv,xh,yh);
  xhidecursor;
  if gpsmarkcircle then
  begin
    drawmarkcircle (x,y);
    errrad := round(gpsestimerr/1852/currentchartsize*(xmax+1));
    if errrad > 8 then
      revcircle (x,y,errrad,round((xasp/yasp)*errrad));
  end
  else
    drawmarkdot (x,y);
  if b then
    revline (x,y,xh,yh);
  xshowcursor;
end;

procedure copygpsdata;
begin
  gpsll.lat := gpslat;
  gpsll.lon := gpslon;
  gpssog := sqrt(sqr(gpsvele)+sqr(gpsveln))*3600/nm;
  if gpssog = 0 then
    gpscog := 0
  else
  begin
    gpscog := calcang(gpsveln,gpsvele);
    if gpscog < 0 then
      gpscog := gpscog+2*pi;
  end;
end;

procedure gpschecklog (draw : boolean);
  var
    x1,y1,x2,y2 : integer;
    tmpcoord : coordptr;
begin
  if not gpslogon then
    exit;
  if gpstime < gpsnextlogtime then
    exit;
  gpsnextlogtime := gpsnextlogtime + gpslogtime;
  { to avoid midnight-problems }
  if gpsnextlogtime > daysec-gpslogtime then
    gpsnextlogtime := gpsnextlogtime - daysec;
  new (tmpcoord);
  insertcoord (tmpcoord,gpsloglastcoord);
  filemod[flog] := true;
  tmpcoord^.ll := gpsll;
  tmpcoord^.coordtype := clineend;
  tmpcoord^.name := gpsloglastcoord^.name;
  gpsloglastcoord^.coordtype := cline;
  if draw then
    draw := round2xy (gpsloglastcoord^.uv,x1,y1);
  gpsloglastcoord := tmpcoord;
  ll2xy (gpsloglastcoord^.ll,gpsloglastcoord^.xy);
  xy2uv (gpsloglastcoord^.xy,gpsloglastcoord^.uv);
  if draw then
    draw := round2xy (gpsloglastcoord^.uv,x2,y2);
  if draw then
  begin
    xhidecursor;
    line (x1,y1,x2,y2);
    xshowcursor;
  end;
end;

function getmenufp (var m : menut) : char;
  var
    i : integer;
    t1,t2,t3,t4 : word;
    r1,r2,r3,r4 : real;
    s : string;
    newdata : boolean;
begin
  drawgpsmark;
  startfine;
  repeat;
    newactpos;
    if showgps then
    begin
      calcdistbear (gpsll,ll,gpsdist,gpsbrg);
      newgpscur;
    end;
    if gpson then
    begin
      repeat
        repeat
        until keypressed or readserok;
        if readserok then
        begin
          if gpsnmea then
          begin
            i := readnmea (s);
            if i = 0 then
            begin
              gpsmsg := 'gps wait';
              if showgps then
                newgpsdata (true);
              drawgpsmark;
              gpsmarktoggle := true;
              drawgpsmark;
            end
            else if i < 0 then
            begin
              gpsmsg := 'gps err';
              if showgps then
                newgpsdata (true);
              drawgpsmark;
              gpsmarktoggle := true;
              drawgpsmark;
            end
            else   { i > 0 }
            begin
              i := getnmeatype(s);
              newdata := false;
              if i <> gpsmsglist[gpsmsgidx] then
                i := 0
              else
              begin
                inc (gpsmsgidx);
                if gpsmsgidx > gpsmsgnum then
                  gpsmsgidx := 1;
              end;
              if i = 1 then
              begin
                if not getnmeagll (s,r1,r2) then
                begin
                  gpsmsg := 'gps err';
                  i := -1;
                end
                else
                begin
                  drawgpsmark;
                  gpslat := r1;
                  gpslon := r2;
                  gpsll.lat := gpslat;
                  gpsll.lon := gpslon;
                  newdata := true;
                end;
              end
              else if i = 2 then
              begin
                if not getnmeavtg (s,r1,r2) then
                begin
                  gpsmsg := 'gps err';
                  i := -1;
                  newdata := true;
                end
                else
                begin
                  drawgpsmark;
                  gpscog := r1;
                  gpssog := r2;
                  newdata := true;
                end;
              end
              else if i = 5 then
              begin
                if not getnmearmc (s,r1,r2,r3,r4) then
                begin
                  gpsmsg := 'gps err';
                  i := -1;
                  newdata := true;
                end
                else
                begin
                  drawgpsmark;
                  gpslat := r1;
                  gpslon := r2;
                  gpsll.lat := gpslat;
                  gpsll.lon := gpslon;
                  gpscog := r3;
                  gpssog := r4;
                  newdata := true;
                end;
              end
              else if i = 6 then
              begin
                if not getnmearme (s,r1) then
                begin
                  gpsmsg := 'gps err';
                  i := -1;
                end
                else
                begin
                  drawgpsmark;
                  gpsestimerr := r1;
                end;
              end;
              if (i = 1) or (i = 5) then
              begin
                getdate (t1,t2,t3,t4);
                gpsdatey := t1;
                gpsdatem := t2;
                gpsdated := t3;
                gettime (t1,t2,t3,t4);
                gpstimeh := t1;
                gpstimem := t2;
                gpstimes := t3;
                gpstime := gpstimeh;
                gpstime := gpstime*60+gpstimem;
                gpstime := gpstime*60+gpstimes;
                gpschecklog (true);
              end;
              if (i = 1) or (i = 2) or (i = 5) or (i = 6) then
              begin
                if gpslogon then
                  gpsmsg := 'gps log '+gpsmsgchar[i]
                else
                  gpsmsg := 'gps ok '+gpsmsgchar[i];
                gpsmarktoggle := not gpsmarktoggle;
                drawgpsmark;
              end
              else if i = -1 then
                if not gpsmarktoggle then
                begin
                  drawgpsmark;
                  gpsmarktoggle := true;
                  drawgpsmark;
                end;
              if (i <> 0) and showgps then
              begin
                if (i = 1) or (i = 5) then
                  calcdistbear (gpsll,ll,gpsdist,gpsbrg);
                newgpsdata (not newdata);
              end;
            end;
          end
          else   { gpspac }
          begin
            gpsbytesleft := readgps;
            if gpsbytesleft = 0 then
            begin
              if gpslogon then
                gpsmsg := 'gps log'
              else
                gpsmsg := 'gps ok';
              getgpsdata;
              drawgpsmark;
              copygpsdata;
              gpschecklog (true);
              gpsmarktoggle := not gpsmarktoggle;
              drawgpsmark;
            end
            else
            begin
              gpsmsg := 'gps err';
              drawgpsmark;
              gpsmarktoggle := true;
              drawgpsmark;
            end;
            if showgps  then
            begin
              calcdistbear (gpsll,ll,gpsdist,gpsbrg);
              newgpsdata (false);
            end;
          end;
        end;
      until keypressed;
    end
    else
      delay (mousedel);
    updatefine (true);
  until anyclick;
  newactpos;
  endfine;
  drawgpsmark;
  getmenufp := checkmenu (m);
end;

procedure drawbigcross;
  var
    ax,ay : integer;
begin
  xhidecursor;
  revline (mx,0,mx,ymax);
  revline (0,my,xmax,my);
  xshowcursor;
  ax := mx;
  ay := my;
  startfine;
  delay (mousedel);
  updatefine (true);
  endfine;
  mkey := false;
  xhidecursor;
  revline (ax,0,ax,ymax);
  revline (0,ay,xmax,ay);
  xshowcursor;
end;

procedure fileformat;
  var
    s1,s2 : string;
    a1 : char;
    a2 : string;
    i,er : integer;
    ok : boolean;
begin
  openwindow;
  newtext (winxc,winyr[1],true,'file format lat&lon');
  newtext (winx1+twb,winyr[6],false,'d -> 12.34567');
  newtext (winx1+twb,winyr[7],false,'m -> 1234.567''');
  newtext (winx1+twb,winyr[8],false,'s -> 1234''56.7"');
  s1 := 'file format [d,m,s]: ';
  s2 := 'lat&lon length     : ';
  a1 := ' ';
  case fllformat of
    1 : a1 := 'd';
    2 : a1 := 'm';
    3 : a1 := 's';
  end;
  str(flllength:2,a2);
  newtext (winx1+twb,winyr[3],false,s1+a1);
  newtext (winx1+twb,winyr[4],false,s2+a2);
  repeat
    if not editchar (winx1+twb,winyr[3],s1,a1) then
    begin
      closewindow;
      exit;
    end;
    ok := (a1 in ['d','m','s']);
    if not ok then
      errorbeep;
  until ok;
  case a1 of
    'd' : fllformat := 1;
    'm' : fllformat := 2;
    's' : fllformat := 3;
  end;
  repeat
    if not editline (winx1+twb,winyr[4],s2,a2) then
    begin
      closewindow;
      exit;
    end;
    val (a2,i,er);
    if er <> 0 then
      errorbeep;
  until er = 0;
  str (i:2,a2);
  newtext (winx1+twb,winyr[4],false,s2+a2+'     ');
  flllength := i;
  closewindow;
end;

procedure inittranscalc (xyar : xyart);
  var
    i : integer;
    ss,si : string;
    b : boolean;
    uvar : uvart;
begin
  if (xyar[2].x-xyar[1].x)*(xyar[3].y-xyar[1].y) -
     (xyar[3].x-xyar[1].x)*(xyar[2].y-xyar[1].y) = 0 then
  begin
    errorbeep;
    b := confirm ('Error: points equal or on one line',false);
    exit;
  end;
  redraw := true;
  for i := 1 to 3 do
  begin
    newscreen;
    str (i,si);
    ss := 'click at point '+si+': '+ll2str(llar[i]);
    boxtext (0,0,textwidth(ss)+tx1+tx2+5,th+ty1+ty2+5,false,ss);
    startfine;
    repeat
      delay (mousedel);
      updatefine (true);
    until anyclick;
    endfine;
    if not clickedlkey then
      exit;
    uvar[i].u := cx;
    uvar[i].v := cy;
  end;
  if not findab (xyar,uvar) then
  begin
    errorbeep;
    b := confirm ('Error: can''t init transformation',false);
  end
  else
  begin
    bmpdisp := false;
    bmpdisplast := false;
    saveab;
    setchartsize;
    recalc := true;
  end;
end;

procedure inittransdigit;
  var
    i : integer;
    si : string;
    so,soll : string;
    s,sll : array [1..3] of string;
    b : boolean;
    lloffar : llart;
    xyar : xyart;
begin
  openwindow;
  newtext (winxc,winyr[1],true,'init transformation');
  so := 'offset (chart datum - WGS84): ';
  soll := ll2str(lloff);
  newtext (winx1+twb,winyr[3],false,so+soll);
  for i := 1 to 3 do
  begin
    lloffar[i].lat := llar[i].lat+lloff.lat;
    lloffar[i].lon := llar[i].lon+lloff.lon;
    sll[i] := ll2str(lloffar[i]);
    str (i,si);
    s[i] := 'coordinates of point '+si+':       ';
    newtext (winx1+twb,winyr[i+3],false,s[i]+sll[i]);
  end;
  repeat
    if not editline (winx1+twb,winyr[3],so,soll) then
    begin
      closewindow;
      exit;
    end;
    b := str2ll(soll,lloff);
    if not b then
      errorbeep;
  until b;
  soll := ll2str(lloff);
  newtext (winx1+twb,winyr[3],false,so+soll+'     ');
  for i := 1 to 3 do
  begin
    lloffar[i].lat := llar[i].lat+lloff.lat;
    lloffar[i].lon := llar[i].lon+lloff.lon;
    sll[i] := ll2str(lloffar[i]);
    newtext (winx1+twb,winyr[i+3],false,s[i]+sll[i]+'     ');
  end;
  for i := 1 to 3 do
  begin
    repeat
      if not editline (winx1+twb,winyr[i+3],s[i],sll[i]) then
      begin
        closewindow;
        exit;
      end;
      b := str2ll(sll[i],lloffar[i]);
      if not b then
        errorbeep;
    until b;
    sll[i] := ll2str(lloffar[i]);
    newtext (winx1+twb,winyr[i+3],false,s[i]+sll[i]+'     ');
    llar[i].lat := lloffar[i].lat-lloff.lat;
    llar[i].lon := lloffar[i].lon-lloff.lon;
    ll2xy (llar[i],xyar[i]);
  end;
  closewindow;
  inittranscalc (xyar);
end;

procedure inittransalign;
  var
    i : integer;
    si : string;
    uv : uvt;
    xyar : xyart;
begin
  redraw := true;
  drawchart;
  initactpos;
  top3x := top4x-textwidth('point 1 | abort')-6-tx1-tx2;
  for i := 1 to 3 do
  begin
    str (i,si);
    showmode ('point '+si+' | abort');
    startfine;
    repeat
      newactpos;
      delay (mousedel);
      updatefine (true);
    until anyclick;
    newactpos;
    endfine;
    if not clickedlkey then
      exit;
    uv.u := cx;
    uv.v := cy;
    uv2xy (uv,xyar[i]);
    xy2ll (xyar[i],llar[i]);
  end;
  inittranscalc (xyar);
end;

procedure inittranschart;
  var
    s1,s2,ans1,ans2 : string;
    b : boolean;
    tempi : integer;
    tempr : real;
begin
  openwindow;
  newtext (winxc,winyr[1],true,'set chart center & scale');
  s1 := 'coordinates of chart center: ';
  ans1 := ll2str(chartcenter);
  s2 := 'width of chart [nm]        : ';
  str (chartwidth:6:4,ans2);
  newtext (winx1+twb,winyr[3],false,s1+ans1);
  newtext (winx1+twb,winyr[4],false,s2+ans2);
  repeat
    if not editline (winx1+twb,winyr[3],s1,ans1) then
    begin
      closewindow;
      exit;
    end;
    b := str2ll(ans1,chartcenter);
    if not b then
      errorbeep;
  until b;
  ans1 := ll2str(chartcenter);
  newtext (winx1+twb,winyr[3],false,s1+ans1+'     ');
  repeat
    if not editline (winx1+twb,winyr[4],s2,ans2) then
    begin
      closewindow;
      exit;
    end;
    val (ans2,tempr,tempi);
    if tempi = 0 then
      chartwidth := tempr
    else
      errorbeep;
  until tempi = 0;
  closewindow;
  chartfindab (chartcenter,chartwidth);
end;

procedure inittransmercator;
begin
  chartwidth := currentchartsize;
  chartcenter := currentchartcenter;
  chartfindab (chartcenter,chartwidth);
end;

procedure zoommouse;
  var
    centx,centy,dx,dy : real;
    x1,y1,x2,y2 : integer;
    uv : uvt;
    xyar : xyart;
    uvar : uvart;
    b : boolean;
    s : string;
begin
  redraw := true;
  drawchart;
  initactpos;
  s := 'center | abort';
  top3x := top4x-textwidth(s)-6-tx1-tx2;
  showmode (s);
  startfine;
  repeat
    newactpos;
    delay (mousedel);
    updatefine (true);
  until anyclick;
  newactpos;
  endfine;
  if not clickedlkey then
    exit;
  centx := cx;
  centy := cy;
  showmode ('size | abort');
  startfine;
  repeat
    dx := abs(cx-centx);
    dy := abs(cy-centy);
    if dx/(xmax+1) > dy/(ymax+1) then
      dy := dx/(xmax+1)*(ymax+1)
    else
      dx := dy/(ymax+1)*(xmax+1);
    x1 := round(centx-dx);
    y1 := round(centy-dy);
    x2 := round(centx+dx);
    y2 := round(centy+dy);
    newactpos;
    xhidecursor;
    revbox (x1,y1,x2,y2);
    xshowcursor;
    delay (mousedel);
    updatefine (true);
    xhidecursor;
    revbox (x1,y1,x2,y2);
    xshowcursor;
  until anyclick;
  newactpos;
  endfine;
  if not clickedlkey then
    exit;
  uvar[1].u := (xmax+1)/2;
  uvar[1].v := (ymax+1)/2;
  uvar[2].u := xmax+1;
  uvar[2].v := (ymax+1)/2;
  uvar[3].u := (xmax+1)/2;
  uvar[3].v := ymax+1;
  uv.u := centx;
  uv.v := centy;
  uv2xy (uv,xyar[1]);
  uv.u := centx+dx;
  uv.v := centy;
  uv2xy (uv,xyar[2]);
  uv.u := centx;
  uv.v := centy+dy;
  uv2xy (uv,xyar[3]);
  if not findab (xyar,uvar) then
  begin
    errorbeep;
    b := confirm ('Error: can''t init transformation',false);
  end
  else
  begin
    setchartsize;
    recalc := true;
    bmpdisp := false;
  end;
end;

procedure zoomfact (fact : real);
  var
    uv : uvt;
    xyar : xyart;
    uvar : uvart;
    b : boolean;
begin
  uvar[1].u := (xmax+1)/2;
  uvar[1].v := (ymax+1)/2;
  uvar[2].u := xmax+1;
  uvar[2].v := (ymax+1)/2;
  uvar[3].u := (xmax+1)/2;
  uvar[3].v := ymax+1;
  uv.u := (xmax+1)/2;
  uv.v := (ymax+1)/2;
  uv2xy (uv,xyar[1]);
  uv.u := (xmax+1)/2*(1+fact);
  uv.v := (ymax+1)/2;
  uv2xy (uv,xyar[2]);
  uv.u := (xmax+1)/2;
  uv.v := (ymax+1)/2*(1+fact);
  uv2xy (uv,xyar[3]);
  if not findab (xyar,uvar) then
  begin
    errorbeep;
    b := confirm ('Error: can''t init transformation',false);
  end
  else
  begin
    setchartsize;
    recalc := true;
    bmpdisp := false;
  end;
end;

procedure transmove (ll : llt);
  var
    xy : xyt;
    uv : uvt;
    xyar : xyart;
    uvar : uvart;
    r : real;
    b : boolean;
begin
  uvar[1].u := (xmax+1)/2;
  uvar[1].v := (ymax+1)/2;
  uvar[2].u := xmax+1;
  uvar[2].v := (ymax+1)/2;
  uvar[3].u := (xmax+1)/2;
  uvar[3].v := ymax+1;
  ll2xy (ll,xy);
  xy2uv (xy,uv);
  uv2xy (uv,xyar[1]);
  r := uv.u;
  uv.u := uv.u + (xmax+1)/2;
  uv2xy (uv,xyar[2]);
  uv.u := r;
  uv.v := uv.v + (ymax+1)/2;
  uv2xy (uv,xyar[3]);
  if not findab (xyar,uvar) then
  begin
    errorbeep;
    b := confirm ('Error: can''t init transformation',false);
  end
  else
  begin
    setchartsize;
    recalc := true;
    bmpdisp := false;
  end;
end;

procedure displlformat;
  var
    s : string;
    c : char;
    i : integer;
    ok : boolean;
begin
  openwindow;
  newtext (winxc,winyr[1],true,'display format lat&lon');
  newtext (winx1+twb,winyr[5],false,'d -> 12.34567');
  newtext (winx1+twb,winyr[6],false,'m -> 1234.567''');
  newtext (winx1+twb,winyr[7],false,'s -> 1234''56.7"');
  c := ' ';
  case llformat of
    1 : c := 'd';
    2 : c := 'm';
    3 : c := 's';
  end;
  s := 'display format [d,m,s]: ';
  repeat
    if not editchar (winx1+twb,winyr[3],s,c) then
    begin
      closewindow;
      exit;
    end;
    ok := (c in ['d','m','s']);
    if not ok then
      errorbeep;
  until ok;
  case c of
    'd' : llformat := 1;
    'm' : llformat := 2;
    's' : llformat := 3;
  end;
  initll2str (llformat,lllength);
  closewindow;
end;

procedure dispgps;
  var
    s1,s2,s3,s4 : string;
    a1,a2,a3 : char;
    a4 : string;
    r : real;
    er : integer;
    ok : boolean;
    oldshowgps : boolean;
begin
  openwindow;
  oldshowgps := showgps;
  newtext (winxc,winyr[1],true,'gps display');
  s1 := 'show gps data line [y/n]:  ';
  s2 := 'show gps mark [y/n]:       ';
  s3 := 'gps mark circle [y/n]:     ';
  s4 := 'headline length [minutes]: ';
  a1 := b2c(showgps);
  a2 := b2c(gpsmark);
  a3 := b2c(gpsmarkcircle);
  str (gpsheadtime:3:1,a4);
  newtext (winx1+twb,winyr[3],false,s1+a1);
  newtext (winx1+twb,winyr[4],false,s2+a2);
  newtext (winx1+twb,winyr[5],false,s3+a3);
  newtext (winx1+twb,winyr[6],false,s4+a4);
  repeat
    if not editchar (winx1+twb,winyr[3],s1,a1) then
    begin
      closewindow;
      exit;
    end;
    ok := c2b(a1,showgps);
    if not ok then
      errorbeep;
  until ok;
  if oldshowgps and (not showgps) then
    redraw := true;
  repeat
    if not editchar (winx1+twb,winyr[4],s2,a2) then
    begin
      closewindow;
      exit;
    end;
    ok := c2b(a2,gpsmark);
    if not ok then
      errorbeep;
  until ok;
  repeat
    if not editchar (winx1+twb,winyr[5],s3,a3) then
    begin
      closewindow;
      exit;
    end;
    ok := c2b(a3,gpsmarkcircle);
    if not ok then
      errorbeep;
  until ok;
  repeat
    if not editline (winx1+twb,winyr[6],s4,a4) then
    begin
      closewindow;
      exit;
    end;
    val (a4,r,er);
    if er <> 0 then
      errorbeep;
  until er = 0;
  gpsheadtime := r;
  closewindow;
end;

procedure dispdrawgrid;
  var
    s,ans : string;
    r : real;
    er : integer;
begin
  openwindow;
  newtext (winxc,winyr[1],true,'grid line distance');
  str (griddist:6:4,ans);
  s := 'distance (0=off) [minutes]: ';
  repeat
    if not editline (winx1+twb,winyr[(winrow+1) div 2],s,ans) then
    begin
      closewindow;
      exit;
    end;
    val (ans,r,er);
    if er = 0 then
    begin
      if (r = 0) or (r >= currentchartsize/maxgridlines) then
        griddist := r
      else
        er := 1;
    end;
    if er <> 0 then
      errorbeep;
  until er = 0;
  closewindow;
  redraw := true;
end;

procedure dispdisplayed;
  var
    s1,s2,s3,s4 : string;
    c1,c2,c3,c4 : char;
    ok : boolean;
begin
  openwindow;
  newtext (winxc,winyr[1],true,'displayed items');
  c1 := b2c(displayvector);
  c2 := b2c(displaytext);
  c3 := b2c(displayovl);
  c4 := b2c(displayscalelimit);
  s1 := 'display vector graphic [y/n]: ';
  s2 := 'display text [y/n]:           ';
  s3 := 'display overlay [y/n]:        ';
  s4 := 'use scale limit [y/n]:        ';
  newtext (winx1+twb,winyr[3],false,s1+c1);
  newtext (winx1+twb,winyr[4],false,s2+c2);
  newtext (winx1+twb,winyr[5],false,s3+c3);
  newtext (winx1+twb,winyr[6],false,s4+c4);
  repeat
    if not editchar (winx1+twb,winyr[3],s1,c1) then
    begin
      closewindow;
      exit;
    end;
    ok := c2b(c1,displayvector);
    if not ok then
      errorbeep;
  until ok;
  recalc := true;
  repeat
    if not editchar (winx1+twb,winyr[4],s2,c2) then
    begin
      closewindow;
      exit;
    end;
    ok := c2b(c2,displaytext);
    if not ok then
      errorbeep;
  until ok;
  repeat
    if not editchar (winx1+twb,winyr[5],s3,c3) then
    begin
      closewindow;
      exit;
    end;
    ok := c2b(c3,displayovl);
    if not ok then
      errorbeep;
  until ok;
  repeat
    if not editchar (winx1+twb,winyr[6],s4,c4) then
    begin
      closewindow;
      exit;
    end;
    ok := c2b(c4,displayscalelimit);
    if not ok then
      errorbeep;
  until ok;
  closewindow;
end;

procedure dispmousecursor;
  var
    s1,s2,s3 : string;
    c1,c2,c3 : char;
    ok : boolean;
begin
  openwindow;
  newtext (winxc,winyr[1],true,'mouse & cursor');
  c1 := b2c(cursorbig);
  c2 := b2c(mousefine);
  c3 := b2c(mouseslow);
  s1 := 'cursor big [y/n]: ';
  s2 := 'mouse fine [y/n]: ';
  s3 := 'mouse slow [y/n]: ';
  newtext (winx1+twb,winyr[3],false,s1+c1);
  newtext (winx1+twb,winyr[4],false,s2+c2);
  newtext (winx1+twb,winyr[5],false,s3+c3);
  repeat
    if not editchar (winx1+twb,winyr[3],s1,c1) then
    begin
      closewindow;
      exit;
    end;
    ok := c2b(c1,cursorbig);
    if not ok then
      errorbeep;
  until ok;
  initfine;
  repeat
    if not editchar (winx1+twb,winyr[4],s2,c2) then
    begin
      closewindow;
      exit;
    end;
    ok := c2b(c2,mousefine);
    if not ok then
      errorbeep;
  until ok;
  repeat
    if not editchar (winx1+twb,winyr[5],s3,c3) then
    begin
      closewindow;
      exit;
    end;
    ok := c2b(c3,mouseslow);
    if not ok then
      errorbeep;
  until ok;
  closewindow;
end;

procedure mainshowinfo;
  var
    chaline,chalinept,chapoly,chapolypt,chatext,chamark : integer;
    ovlline,ovllinept,ovlpoly,ovlpolypt,ovltext,ovlmark : integer;
    logline,loglinept,memmark,wptmark : integer;
    actblock : blockptr;
    actcoord : coordptr;
    s,si1,si2 : string;
begin
  chaline := 0;
  chalinept := 0;
  chapoly := 0;
  chapolypt := 0;
  chatext := 0;
  chamark := 0;
  ovlline := 0;
  ovllinept := 0;
  ovlpoly := 0;
  ovlpolypt := 0;
  ovltext := 0;
  ovlmark := 0;
  logline := 0;
  loglinept := 0;
  memmark := 0;
  wptmark := 0;
  actblock := blockfirst;
  while actblock <> nil do
  begin
    actcoord := actblock^.coord;
    case actblock^.filetype of
      fcha : case actcoord^.coordtype of
               cline : begin
                         inc (chaline);
                         inc (chalinept);
                       end;
               cpoly : begin
                         inc (chapoly);
                         inc (chapolypt);
                       end;
               ctext : inc (chatext);
               cmark : inc (chamark);
             end;
      fovl : case actcoord^.coordtype of
               cline : begin
                         inc (ovlline);
                         inc (ovllinept);
                       end;
               cpoly : begin
                         inc (ovlpoly);
                         inc (ovlpolypt);
                       end;
               ctext : inc (ovltext);
               cmark : inc (ovlmark);
             end;
      flog : case actcoord^.coordtype of
               cline : begin
                         inc (logline);
                         inc (loglinept);
                       end;
             end;
      fmem : case actcoord^.coordtype of
               cmark : inc (memmark);
             end;
      fwpt : case actcoord^.coordtype of
               cmark : inc (wptmark);
             end;
    end;
    while actcoord^.coordtype = cline do
    begin
      case actblock^.filetype of
        fcha : inc (chalinept);
        fovl : inc (ovllinept);
        flog : inc (loglinept);
      end;
      actcoord := actcoord^.succ;
    end;
    while actcoord^.coordtype = cpoly do
    begin
      case actblock^.filetype of
        fcha : inc (chapolypt);
        fovl : inc (ovlpolypt);
      end;
      actcoord := actcoord^.succ;
    end;
    actblock := actblock^.succ;
  end;
  openwindow;
  newtext (winxc,winyr[1],true,'data info');
  str (chaline:4,si1);
  str (chalinept:4,si2);
  s := 'chart:    '+si1+' lines ('+si2+' points)  ';
  str (ovlline:4,si1);
  str (ovllinept:4,si2);
  s := s+'overlay:  '+si1+' lines ('+si2+' points)';
  newtext (winx1+twb,winyr[3],false,s);
  str (chapoly:4,si1);
  str (chapolypt:4,si2);
  s := 'chart:    '+si1+' polys ('+si2+' points)  ';
  str (ovlpoly:4,si1);
  str (ovlpolypt:4,si2);
  s := s+'overlay:  '+si1+' polys ('+si2+' points)';
  newtext (winx1+twb,winyr[4],false,s);
  str (chatext:4,si1);
  s := 'chart:    '+si1+' texts                ';
  str (ovltext:4,si1);
  s := s+'overlay:  '+si1+' texts';
  newtext (winx1+twb,winyr[5],false,s);
  str (chamark:4,si1);
  s := 'chart:    '+si1+' marks                ';
  str (ovlmark:4,si1);
  s := s+'overlay:  '+si1+' marks';
  newtext (winx1+twb,winyr[6],false,s);
  str (logline:4,si1);
  str (loglinept:4,si2);
  s := 'log:      '+si1+' lines ('+si2+' points)';
  newtext (winx1+twb,winyr[7],false,s);
  str (memmark:4,si1);
  s := 'memo:     '+si1+' marks';
  newtext (winx1+twb,winyr[8],false,s);
  str (wptmark:4,si1);
  s := 'waypoint: '+si1+' marks                ';
  str (memavail/initialmemavail*100:3:1,si1);
  str (initialmemavail div 1024:1,si2);
  s := s+'memory:   '+si1+'% of '+si2+'K free';
  newtext (winx1+twb,winyr[9],false,s);
  waitclickkey;
  closewindow;
end;

procedure seteditmode (m : editmodet);
begin
  editmode := m;
  showmode (editmodename[editmode]);
end;

procedure showhelp;
begin
  openwindow;
  newtext (winxc,winyr[1],true,'help 1/4');
  newtext (winxc,winyr[3],true,
           title+'   '+version+'   '+author);
  newtext (winxc,winyr[5],true,
           'All usual disclaimers apply! You use CHART at your own risk!');
  newtext (winxc,winyr[7],true,
           'written by Heiko Purnhagen              ');
  newtext (winxc,winyr[8],true,
           'e-mail: purnhage@tnt.uni-hannover.de    ');
  newtext (winxc,winyr[9],true,
           'http://www.fet.uni-hannover.de/purnhage/');
  waitclickkey;
  closewindow;
  openwindow;
  newtext (winxc,winyr[1],true,'help 2/4');
  newtext (winx1+twb,winyr[3],false,
           'mouse key functions displayed in upper right corner:');
  newtext (winx1+twb,winyr[4],false,
           '      left button | right button');
  newtext (winx1+twb,winyr[6],false,
           'keys: <spc>  = left button       <esc> & others = right button');
  newtext (winx1+twb,winyr[7],false,
           '      <pgup> = zoom out          <pgdn>         = zoom in');
  newtext (winx1+twb,winyr[8],false,
           '      <home> = move              <end>          = last scale');
  newtext (winx1+twb,winyr[9],false,
           '      <ins>  = set gps lat&lon   <tab>          = center gps');
  waitclickkey;
  closewindow;
  openwindow;
  newtext (winxc,winyr[1],true,'help 3/4');
  newtext (winx1+twb,winyr[3],false,
           'move mouse cursor    slow    fast');
  newtext (winx1+twb,winyr[5],false,
           '      up             <up>     ^E');
  newtext (winx1+twb,winyr[6],false,
           '     down           <down>    ^X');
  newtext (winx1+twb,winyr[7],false,
           '     left           <left>    ^S');
  newtext (winx1+twb,winyr[8],false,
           '     right          <right>   ^D');
  waitclickkey;
  closewindow;
  openwindow;
  newtext (winxc,winyr[1],true,'help 4/4');
  newtext (winx1+twb,winyr[3],false,
           'rel. text position type char:   a b c');
  newtext (winx1+twb,winyr[4],false,
           '                                d e f');
  newtext (winx1+twb,winyr[5],false,
           '                                g h i');
  newtext (winx1+twb,winyr[7],false,
           'line/poly type char:   a = solid line    b = dotted line');
  newtext (winx1+twb,winyr[8],false,
           '(optional)             c = center line   d = dashed line');
  newtext (winx1+twb,winyr[9],false,
           '                       e = thick solid line');
  waitclickkey;
  closewindow;
end;

function selectfile (filetypename : string; actmask : string;
                     var actfile : string) : boolean;
{ OK->true escape->false }
  const
    maxdir = 100;
  var
    srec : searchrec;
    dir : array[0..maxdir] of string[14];
    numdir : integer;
    s : string;
    c : char;
    actpath : string;
    b : boolean;
    act,i,j : integer;
begin
  actfile := '';
  selectfile := false;
  repeat
    if not editline (winx1+twb,winyr[(winrow+1) div 2],
                     filetypename+' filename: ',actmask) then
      exit;
    if (pos('*',actmask) > 0) or (pos('?',actmask) > 0) then
    begin
      i := length(actmask);
      while (i > 0) and (actmask[i] <> '\') and (actmask[i] <> ':') do
        dec (i);
      actpath := copy(actmask,1,i);
      numdir := 0;
      findfirst (actmask,0,srec);
      while (doserror = 0) and (numdir < maxdir) do
      begin
        inc (numdir);
        dir[numdir] := srec.name;
        findnext (srec);
      end;
      if numdir = 0 then
        b := confirm ('no files found',false)
      else
      begin
        if numdir >= maxdir then
        begin
          str (maxdir,s);
          b := confirm ('more than '+s+' files found',false);
        end;
        { bubblesort }
        for i := numdir-1 downto 1 do
          for j := 1 to i do
            if dir[j] > dir[j+1] then
            begin
              dir[0] := dir[j];
              dir[j] := dir[j+1];
              dir[j+1] := dir[0];
            end;
        act := 1;
        repeat
          newtext (winx1+twb,winyr[(winrow+1) div 2],false,
                   'select '+filetypename+' file: '+dir[act]+
                   '              ');
          repeat
            delay (mousedel);
            updatefine (false);
          until clickedr or mkey;
          i := 0;
          if clickedr then
            i := 1
          else   { mkey }
          begin
            c := getmkey;
            if c = esc then
              i := 1;
            if c = cr then
              i := 2;
            if c = backsp then
              i := 3;
            if (c = ' ') or (c = tab) then
              i := 4;
            if c = #0 then
            begin
              c := mc2;
              if (c = cleft) or (c = cup) then
                i := 3;
              if (c = cright) or (c = cdown) then
                i := 4;
            end;
          end;
          if i = 3 then
          begin
            if act > 1 then
              dec (act)
            else
              errorbeep;
          end;
          if i = 4 then
          begin
            if act < numdir then
              inc (act)
            else
              errorbeep;
          end;
        until (i = 1) or (i = 2);
        if i = 2 then
          actfile := actpath+dir[act];
        newtext (winx1+twb,winyr[(winrow+1) div 2],false,
                 '       '+'        '+'       '+'              ');
      end;
    end
    else
      actfile := actmask;
  until actfile <> '';
  selectfile := true;
end;

function loadsave (filetype : filetypet; doload : boolean) : boolean;
  var
    actmask,actfile : string;
    s : string;
    b : boolean;
begin
  loadsave := false;
  if (flllength < lllength) and not doload then
  begin
    str (flllength,s);
    if not confirm('use lat&lon length '+s+' (-> low resolution)?',true) then
      exit;
  end;
  if doload then
    actmask := filemask[filetype]
  else
    actmask := filename[filetype];
  if not selectfile(filetypename[filetype],actmask,actfile) then
    exit;
  filename[filetype] := actfile;
  if doload then
  begin
    newtext (winx1+twb,winyr[(winrow+1) div 2],false,
             'loading '+filetypename[filetype]+' file '+filename[filetype]+' ... ');
    b := loadd (filename[filetype],filetype);
  end
  else
  begin
    newtext (winx1+twb,winyr[(winrow+1) div 2],false,
             'saving '+filetypename[filetype]+' file '+filename[filetype]+' ... ');
    if saved (filename[filetype],filetype) then
      filemod[filetype] := false;
  end;
  loadsave := true;
end;

procedure filebmpalign;
  var
    f : file;
    wid,hig,col : integer;
    c00 : boolean;
    xcenter,ycenter : real;
    thr : integer;
    er : integer;
    i : integer;
    ss,st,si,so,soll : string;
    s,sll : array [1..3] of string;
    b : boolean;
    ok : boolean;
    lloffar : llart;
    xyar : xyart;
    uvar : uvart;
    tmpfact : real;
    tmpxoff,tmpyoff,tmpx0,tmpy0 : integer;
    actmask,actfile : string;
    bmpname,bmcname : string;
    bgc,fgc : integer;
begin
  openwindow;
  newtext (winxc,winyr[1],true,'align bitmap');
  if not selectfile('bitmap','*.bmp',actfile) then
  begin
    closewindow;
    exit;
  end;
  bmpname := actfile;
  ss := readbmphdr(bmpname,wid,hig,col,c00);
  if ss <> '' then
  begin
    b := confirm('Error: '+bmpname+' '+ss,false);
    closewindow;
    exit;
  end;
  closewindow;
  openwindow;
  thr := 1;
  newtext (winxc,winyr[1],true,'align bitmap');
  newtext (winx1+twb,winyr[3],false,'bitmap file:                  '+bmpname);
  str (wid,s[1]);
  str (hig,s[2]);
  str (col,s[3]);
  newtext (winx1+twb,winyr[4],false,'                              '+
           '(width='+s[1]+' height='+s[2]+'  '+s[3]+' colors)');
  ss := 'threshold (<0 -> reverse):    ';
  str (thr,st);
  newtext (winx1+twb,winyr[5],false,ss+st);
  so := 'offset (chart datum - WGS84): ';
  soll := ll2str(lloff);
  newtext (winx1+twb,winyr[6],false,so+soll);
  for i := 1 to 3 do
  begin
    lloffar[i].lat := llar[i].lat+lloff.lat;
    lloffar[i].lon := llar[i].lon+lloff.lon;
    sll[i] := ll2str(lloffar[i]);
    str (i,si);
    s[i] := 'coordinates of point '+si+':       ';
    newtext (winx1+twb,winyr[i+6],false,s[i]+sll[i]);
  end;
  repeat
    if not editline (winx1+twb,winyr[5],ss,st) then
    begin
      closewindow;
      exit;
    end;
    val (st,thr,er);
    if er <> 0 then
      errorbeep;
  until er = 0;
  str (thr,st);
  newtext (winx1+twb,winyr[5],false,ss+st+'     ');
  repeat
    if not editline (winx1+twb,winyr[6],so,soll) then
    begin
      closewindow;
      exit;
    end;
    b := str2ll(soll,lloff);
    if not b then
      errorbeep;
  until b;
  soll := ll2str(lloff);
  newtext (winx1+twb,winyr[6],false,so+soll+'     ');
  for i := 1 to 3 do
  begin
    lloffar[i].lat := llar[i].lat+lloff.lat;
    lloffar[i].lon := llar[i].lon+lloff.lon;
    sll[i] := ll2str(lloffar[i]);
    newtext (winx1+twb,winyr[i+6],false,s[i]+sll[i]+'     ');
  end;
  for i := 1 to 3 do
  begin
    repeat
      if not editline (winx1+twb,winyr[i+6],s[i],sll[i]) then
      begin
        closewindow;
        exit;
      end;
      b := str2ll(sll[i],lloffar[i]);
      if not b then
        errorbeep;
    until b;
    sll[i] := ll2str(lloffar[i]);
    newtext (winx1+twb,winyr[i+6],false,s[i]+sll[i]+'     ');
    llar[i].lat := lloffar[i].lat-lloff.lat;
    llar[i].lon := lloffar[i].lon-lloff.lon;
    ll2xy (llar[i],xyar[i]);
  end;
  closewindow;
  if (xyar[2].x-xyar[1].x)*(xyar[3].y-xyar[1].y) -
     (xyar[3].x-xyar[1].x)*(xyar[2].y-xyar[1].y) = 0 then
  begin
    errorbeep;
    b := confirm ('Error: points equal or on one line',false);
    exit;
  end;
  redraw := true;
  bmpdisp := false;
  bmpdisplast := false;
  if bmpbuf[0] <> nil then
  begin
    if not loadbitmap (bmpname,wid div 2,hig div 2,
                       xmax+1,ymax+1,-999,thr,getmaxcolor,true) then
      exit;
    storebitmap;
    tmpxoff := bmpxoff;
    tmpyoff := bmpyoff;
    tmpx0 := bmpx0;
    tmpy0 := bmpy0;
    tmpfact := bmpfact;
  end;
  for i := 1 to 3 do
  begin
    repeat
      if bmpbuf[0] = nil then
      begin
        if not loadbitmap (bmpname,wid div 2,hig div 2,
                           xmax+1,ymax+1,-999,thr,getmaxcolor,true) then
          exit;
      end
      else
      begin
        restorebitmap;
        bmpxoff := tmpxoff;
        bmpyoff := tmpyoff;
        bmpx0 := tmpx0;
        bmpy0 := tmpy0;
        bmpfact := tmpfact;
      end;
      str (i,si);
      ss := 'click at point '+si+': '+ll2str(llar[i]);
      boxtext (0,0,textwidth(ss)+tx1+tx2+5,th+ty1+ty2+5,false,ss);
      startfine;
      repeat
        delay (mousedel);
        updatefine (true);
      until anyclick;
      endfine;
      if not clickedlkey then
        exit;
      calcbitmaps2b (cx,cy,xcenter,ycenter);
      if not loadbitmap (bmpname,round(xcenter),round(ycenter),
                         (xmax+1) div 2,(ymax+1) div 2,0,
                         thr,getmaxcolor,true) then
        exit;
      str (i,si);
      ss := 'click at point '+si+': '+ll2str(llar[i]);
      boxtext (0,0,textwidth(ss)+tx1+tx2+5,th+ty1+ty2+5,false,ss);
      startfine;
      repeat
        delay (mousedel);
        updatefine (true);
      until anyclick;
      endfine;
      ok := clickedlkey;
      calcbitmaps2b (cx,cy,uvar[i].u,uvar[i].v);
    until ok;
  end;
  newscreen;
  saveabtmp;
  if not findab (xyar,uvar) then
  begin
    errorbeep;
    b := confirm ('Error: can''t init transformation',false);
  end
  else
  begin
    i := pos('.',bmpname);
    if i > 0 then
      actmask := copy(bmpname,1,i)+'bmc'
    else
      actmask := bmpname+'.bmc';
    bmcname := '';
    openwindow;
    repeat
      newtext (winxc,winyr[1],true,'save bitmap chart');
      if not selectfile ('bitmap chart',actmask,actfile) then
      begin
        closewindow;
        exit;
      end;
      assign (f,actfile);
      {$i-}
      reset (f);
      {$i+}
      if ioresult = 0 then
      begin
        close (f);
        if confirm (actfile+' exists! overwrite?',true) then
          bmcname := actfile;
      end
      else
        bmcname := actfile;
    until bmcname <> '';
    closewindow;
    b := savebmc (bmcname,bmpname,thr,llar,uvar);
  end;
  restoreabtmp;
end;

procedure filebmpload;
  var
    f : file;
    s : string;
    b : boolean;
    i : integer;
    s1,s2,s3,a1,a2,a3 : string;
    er : integer;
    actmask,actfile : string;
    bmpname : string;
    bmcthr : integer;
    bmcllar : llart;
    xyar : xyart;
    uvar : uvart;
    wid,hig,col : integer;
    c00 : boolean;
begin
  openwindow;
  newtext (winxc,winyr[1],true,'load bitmap chart');
  if not selectfile('bitmap chart','*.bmc',actfile) then
  begin
    closewindow;
    exit;
  end;
  bmcnameload := actfile;
  if not loadbmc(bmcnameload,bmpname,bmcthr,bmcllar,uvar) then
  begin
    closewindow;
    exit;
  end;
  assign (f,bmpname);
  {$i-}
  reset (f);
  {$i+}
  if ioresult <> 0 then
  begin
    i := length(bmpname);
    while (i > 0) and (bmpname[i] <> '\') and (bmpname[i] <> ':') do
      dec (i);
    s := copy(bmpname,i+1,length(bmpname)-i);
    i := length(bmcnameload);
    while (i > 0) and (bmcnameload[i] <> '\') and (bmcnameload[i] <> ':') do
      dec (i);
    bmpname := copy(bmcnameload,1,i)+s;
    assign (f,bmpname);
    {$i-}
    reset (f);
    {$i+}
    if ioresult <> 0 then
    begin
      b := confirm ('Error: can''t find '+s,false);
      closewindow;
      exit;
    end
  end;
  s := readbmphdr(bmpname,wid,hig,col,c00);
  if s <> '' then
  begin
    b := confirm('Error: '+bmpname+' '+s,false);
    closewindow;
    exit;
  end;
  closewindow;
  bmpnameload := bmpname;
  bmpthr := bmcthr;
  bmpllar := bmcllar;
  llar := bmpllar;
  bmpuvar := uvar;
  bmpcenterll := currentchartcenter;
  bmpdisp := false;
  bmpdisplast := false;
  openwindow;
  newtext (winxc,winyr[1],true,'load bitmap chart');
  newtext (winx1+twb,winyr[3],false,
           'bitmap chart file:              '+bmcnameload);
  newtext (winx1+twb,winyr[4],false,
           'bitmap file:                    '+bmpnameload);
  saveabtmp;
  for i := 1 to 3 do
    ll2xy (bmpllar[i],xyar[i]);
  if not findab (xyar,bmpuvar) then
  begin
    errorbeep;
    b := confirm ('Error: can''t init transformation',false);
    exit;
  end;
  uv.u := 0;
  uv.v := 0;
  uv2xy (uv,xy);
  xy2ll (xy,ll);
  s := ll2str (ll);
  newtext (winx1+twb,winyr[5],false,'lower left corner  (   0,   0): '+s);
  str (wid-1:4,s1);
  str (hig-1:4,s2);
  uv.u := wid-1;
  uv.v := hig-1;
  uv2xy (uv,xy);
  xy2ll (xy,ll);
  s := ll2str (ll);
  newtext (winx1+twb,winyr[6],false,'upper right corner ('+s1+','+s2+'): '+s);
  restoreabtmp;
  s1 := 'coordinates of chart center:    ';
  a1 := ll2str (bmpcenterll);
  s2 := 'zoom (-1->50% 0->100% 1->200%): ';
  str (bmpzoom,a2);
  str (getmaxcolor:2,s);
  s3 := 'color ( 1..'+s+'):                 ';
  str (bmpcol,a3);
  newtext (winx1+twb,winyr[7],false,s1+a1);
  newtext (winx1+twb,winyr[8],false,s2+a2);
  newtext (winx1+twb,winyr[9],false,s3+a3);
  repeat
    if not editline (winx1+twb,winyr[7],s1,a1) then
    begin
      closewindow;
      exit;
    end;
    b := str2ll(a1,bmpcenterll);
    if not b then
      errorbeep;
  until b;
  a1 := ll2str(bmpcenterll);
  newtext (winx1+twb,winyr[7],false,s1+a1+'     ');
  repeat
    if not editline (winx1+twb,winyr[8],s2,a2) then
    begin
      closewindow;
      exit;
    end;
    val (a2,bmpzoom,er);
    if er <> 0 then
      errorbeep;
  until er = 0;
  str (bmpzoom,a2);
  newtext (winx1+twb,winyr[8],false,s2+a2+'     ');
  repeat
    if not editline (winx1+twb,winyr[9],s3,a3) then
    begin
      closewindow;
      exit;
    end;
    val (a3,bmpcol,er);
    if er <> 0 then
      errorbeep
    else
      if (bmpcol < 1) or (bmpcol > getmaxcolor) then
      begin
        er := -1;
        errorbeep;
      end;
  until er = 0;
  str (bmpcol,a3);
  newtext (winx1+twb,winyr[9],false,s3+a3+'     ');
  closewindow;
end;

procedure filebmpreload;
  var
    i : integer;
    b : boolean;
    uvar : uvart;
    xyar : xyart;
begin
  if bmpnameload = '' then
  begin
    b := confirm ('Error: no bitmap chart loaded',false);
    exit;
  end;
  redraw := true;
  bmpdisp := false;
  bmpdisplast := false;
  saveabtmp;
  for i := 1 to 3 do
    ll2xy (bmpllar[i],xyar[i]);
  if not findab (xyar,bmpuvar) then
  begin
    errorbeep;
    b := confirm ('Error: can''t init transformation',false);
    exit;
  end;
  ll2xy (bmpcenterll,xy);
  xy2uv (xy,uv);
  bmpxcenter := round(uv.u);
  bmpycenter := round(uv.v);
  restoreabtmp;
  if loadbitmap (bmpnameload,bmpxcenter,bmpycenter,xmax+1,ymax+1,
                 bmpzoom,bmpthr,bmpcol,bmpbuf[0]<>nil) then
  begin
    storebitmap;
    for i := 1 to 3 do
      calcbitmapb2s (bmpuvar[i].u,bmpuvar[i].v,uvar[i].u,uvar[i].v);
    if not findab (xyar,uvar) then
    begin
      errorbeep;
      b := confirm ('Error: can''t init transformation',false);
      exit;
    end;
    saveab;
    setchartsize;
    recalc := true;
    bmpdisp := true;
    bmpdisplast := true;
  end;
end;

procedure filebmpsavescreen;
  var
    actfile,actmask : string;
    b : boolean;
    f : file;
    s,ss : string;
begin
  openwindow;
  newtext (winxc,winyr[1],true,'save screen bitmap');
  actmask := bmpnamesave;
  if not selectfile ('bitmap',actmask,actfile) then
  begin
    closewindow;
    exit;
  end;
  assign (f,actfile);
  {$i-}
  reset (f);
  {$i+}
  if ioresult = 0 then
  begin
    close (f);
    if not confirm (actfile+' exists! overwrite?',true) then
      exit;
  end;
  bmpnamesave := actfile;
  closewindow;
  drawchart;
  s := ll2str(currentchartcenter);
  newtext (0,0,true,s);
  str (currentchartsize:5:3,ss);
  if length(ss) > 5 then
    str (currentchartsize:5:2,ss);
  if length(ss) > 5 then
    str (currentchartsize:5:1,ss);
  if length(ss) > 5 then
    str (currentchartsize:5:0,ss);
  s := 'center: '+s+'   size: '+ss+'nm';
  boxtext (0,0,xmax,th+ty1+ty2+5,true,s);
  if savebitmap (bmpnamesave) then
    b := confirm ('screen bitmap saved',false);
end;


{ edit procedures }

function findcoord (cx,cy : real) : coordptr;
  var
    actcoord : coordptr;
    mindist,dist : real;
begin
  mindist := sqr(maxfind);
  findcoord := nil;
  actcoord := coordfirst;
  while actcoord <> nil do
  begin
    dist := sqr(actcoord^.uv.u-cx)+sqr(actcoord^.uv.v-cy);
    if dist < mindist then
    begin
      findcoord := actcoord;
      mindist := dist;
    end;
    actcoord := actcoord^.succ;
  end;
end;

function editcoord (actcoord : coordptr; modify,modifypos : boolean) : boolean;
  var
    s,ss : string;
    s1,s2,s3,s4 : string;
    a1,a2,a3 : string;
    a4 : char;
    ll : llt;
    scalelimit : real;
    thispoint,linepoints,r : integer;
    tmpcoord : coordptr;
    endit,ok : boolean;
    er : integer;
begin
  editcoord := false;
  r := 3;
  if actcoord^.coordtype in [cline,clineend,cpoly,cpolyend] then
  begin
    inc (r);
    thispoint := 0;
    linepoints := 0;
    tmpcoord := actcoord^.name^.coord;
    repeat
      inc (linepoints);
      if tmpcoord = actcoord then
        thispoint := linepoints;
      endit := (tmpcoord^.coordtype in [clineend,cpolyend]);
      tmpcoord := tmpcoord^.succ;
    until endit;
    str (thispoint:4,s);
    str (linepoints:4,ss);
    newtext (winx1+twb,winyr[3],false,
             'point nr:        '+s+'   ('+ss+
             ' points in '+coordtypename[actcoord^.coordtype]+')');
  end;
  s1 := 'position:        ';
  s2 := 'name:            ';
  s3 := 'scalelimit [nm]: ';
  s4 := 'type char:       ';
  a1 := ll2str(actcoord^.ll);
  a2 := actcoord^.name^.name;
  str (actcoord^.name^.scalelimit:6:4,a3);
  a4 := actcoord^.name^.typec;
  newtext (winx1+twb,winyr[r],false,s1+a1);
  newtext (winx1+twb,winyr[r+1],false,s2+a2);
  newtext (winx1+twb,winyr[r+2],false,s3+a3);
  newtext (winx1+twb,winyr[r+3],false,s4+a4);
  newtext (winx1+twb,winyr[r+5],false,
           'scalelimit = 0 -> display always');
  if not modify then
    waitclickkey
  else
  begin
    if modifypos then
    begin
      repeat
        if not editline (winx1+twb,winyr[r],s1,a1) then
          exit;
        ok := str2ll(a1,ll);
        if not ok then
          errorbeep;
      until ok;
      a1 := ll2str(ll);
      newtext (winx1+twb,winyr[r],false,s1+a1+'     ');
    end;
    if not editline (winx1+twb,winyr[r+1],s2,a2) then
    begin
      closewindow;
      exit;
    end;
    newtext (winx1+twb,winyr[r+1],false,s2+a2+'     ');
    repeat
      if not editline (winx1+twb,winyr[r+2],s3,a3) then
        exit;
      val (a3,scalelimit,er);
      if er <> 0 then
        errorbeep;
    until er = 0;
    str (scalelimit:6:4,a3);
    newtext (winx1+twb,winyr[r+2],false,s3+a3+'     ');
    repeat
      if not editchar (winx1+twb,winyr[r+3],s4,a4) then
        exit;
      ok := (a4 >= ' ');
      if not ok then
        errorbeep;
    until ok;
    if modifypos then
    begin
      actcoord^.ll := ll;
      ll2xy (actcoord^.ll,actcoord^.xy);
      xy2uv (actcoord^.xy,actcoord^.uv);
    end;
    actcoord^.name^.name := a2;
    actcoord^.name^.scalelimit := scalelimit;
    actcoord^.name^.typec := a4;
    filemod[actcoord^.name^.filetype] := true;
    redraw := true;
  end;
  editcoord := true;
end;

procedure addlinepoly (filetype : filetypet; coordtype : coordtypet);
  var
    lx,ly,lxp,lyp,lxs,lys : integer;
    rubbers,b,place : boolean;
    s,name : string;
    actcoord,startcoord,nextcoord : coordptr;
    block : blockptr;
begin
  showmode ('place | quit');
  new (block);
  new (actcoord);
  startcoord := actcoord;
  actcoord^.name := block;
  actcoord^.ll := ll;
  actcoord^.coordtype := coordtype;
  insertcoord (actcoord,coordlast);
  ll2xy (actcoord^.ll,actcoord^.xy);
  xy2uv (actcoord^.xy,actcoord^.uv);
  b := round2xy (actcoord^.uv,lxp,lyp);
  if coordtype = cpoly then
  begin
    rubbers := true;
    lxs := lxp;
    lys := lyp;
  end
  else
    rubbers := false;
  repeat
    startfine;
    repeat
      lx := round(cx);
      ly := round(cy);
      newactpos;
      xhidecursor;
      revline (lxp,lyp,lx,ly);
      if rubbers and ((lxp <> lxs) or (lyp <> lys)) then
        revline (lx,ly,lxs,lys);
      xshowcursor;
      delay (mousedel);
      updatefine (true);
      xhidecursor;
      revline (lxp,lyp,lx,ly);
      if rubbers and ((lxp <> lxs) or (lyp <> lys)) then
        revline (lx,ly,lxs,lys);
      xshowcursor;
    until anyclick;
    newactpos;
    endfine;
    place := clickedlkey;
    if place then
    begin
      new (actcoord);
      actcoord^.name := block;
      actcoord^.ll := ll;
      actcoord^.coordtype := coordtype;
      insertcoord (actcoord,coordlast);
      ll2xy (actcoord^.ll,actcoord^.xy);
      xy2uv (actcoord^.xy,actcoord^.uv);
      b := round2xy (actcoord^.uv,lx,ly);
      xhidecursor;
      line (lxp,lyp,lx,ly);
      xshowcursor;
      lxp := lx;
      lyp := ly;
    end;
  until not place;
  if coordtype = cline then
    actcoord^.coordtype := clineend
  else
    actcoord^.coordtype := cpolyend;
  if coordtype = cpoly then
  begin
    xhidecursor;
    line (lxp,lyp,lxs,lys);
    xshowcursor;
  end;
  openwindow;
  newtext (winxc,winyr[1],true,'add '+filetypename[filetype]+' '+
           coordtypename[coordtype]);
  block^.coord := startcoord;
  block^.name := '';
  block^.filetype := filetype;
  block^.scalelimit := defscalelimit;
  block^.typec := ' ';
  if not editcoord (startcoord,true,false) then
  begin
    actcoord := startcoord;
    while actcoord <> nil do
    begin
      nextcoord := actcoord^.succ;
      removecoord (actcoord);
      dispose (actcoord);
      actcoord := nextcoord;
    end;
    dispose (block);
    closewindow;
    exit;
  end;
  defscalelimit := block^.scalelimit;
  insertblock (block,blocklast);
  filemod[filetype] := true;
  redraw := true;
  closewindow;
end;

procedure addmarktext (filetype : filetypet; coordtype : coordtypet);
  var
    name : string;
    coord : coordptr;
    block : blockptr;
begin
  openwindow;
  newtext (winxc,winyr[1],true,'add '+filetypename[filetype]+' '+
           coordtypename[coordtype]);
  new (coord);
  new (block);
  coord^.name := block;
  coord^.ll := ll;
  ll2xy (coord^.ll,coord^.xy);
  xy2uv (coord^.xy,coord^.uv);
  coord^.coordtype := coordtype;
  block^.coord := coord;
  block^.name := '';
  block^.filetype := filetype;
  block^.scalelimit := defscalelimit;
  block^.typec := ' ';
  if not editcoord (coord,true,false) then
  begin
    dispose (block);
    dispose (coord);
    closewindow;
    exit;
  end;
  defscalelimit := block^.scalelimit;
  insertcoord (coord,coordlast);
  insertblock (block,blocklast);
  filemod[filetype] := true;
  redraw := true;
  closewindow;
end;

procedure editmovelinepoly (actcoord : coordptr);
  var
    lx,ly,lxp,lyp,lxs,lys : integer;
    rubberp,rubbers,b : boolean;
    tmpcoord : coordptr;
begin
  showmode ('place | abort');
  b := round2xy (actcoord^.uv,lx,ly);
  if actcoord^.name^.coord <> actcoord then
  begin
    rubberp := true;
    tmpcoord := actcoord^.pred;
  end
  else
    if actcoord^.coordtype = cpoly then
    begin
      rubberp := true;
      tmpcoord := actcoord;
      while tmpcoord^.coordtype = cpoly do
        tmpcoord := tmpcoord^.succ;
    end
    else
      rubberp := false;
  if rubberp then
    rubberp := round2xy (tmpcoord^.uv,lxp,lyp);
  if actcoord^.coordtype in [cline,cpoly] then
  begin
    rubbers := true;
    tmpcoord := actcoord^.succ;
  end
  else
    if actcoord^.coordtype = cpolyend then
    begin
      rubbers := true;
      tmpcoord := actcoord^.name^.coord;
    end
    else
      rubbers := false;
  if rubbers then
    rubbers := round2xy (tmpcoord^.uv,lxs,lys);
  setcolor (bgc);
  xhidecursor;
  if rubberp then
    line (lxp,lyp,lx,ly);
  if rubbers and ((not rubberp) or (lxp <> lxs) or (lyp <> lys)) then
    line (lx,ly,lxs,lys);
  xshowcursor;
  setcolor (fgc);
  redraw := true;
  startfine;
  repeat
    lx := round(cx);
    ly := round(cy);
    newactpos;
    xhidecursor;
    if rubberp then
      revline (lxp,lyp,lx,ly);
    if rubbers and ((not rubberp) or (lxp <> lxs) or (lyp <> lys)) then
      revline (lx,ly,lxs,lys);
    xshowcursor;
    delay (mousedel);
    updatefine (true);
    xhidecursor;
    if rubberp then
      revline (lxp,lyp,lx,ly);
    if rubbers and ((not rubberp) or (lxp <> lxs) or (lyp <> lys)) then
      revline (lx,ly,lxs,lys);
    xshowcursor;
  until anyclick;
  newactpos;
  endfine;
  if clickedlkey then
  begin
    actcoord^.ll := ll;
    ll2xy (actcoord^.ll,actcoord^.xy);
    xy2uv (actcoord^.xy,actcoord^.uv);
    filemod[actcoord^.name^.filetype] := true;
    redraw := true;
  end;
end;

procedure editmovemarktext (actcoord : coordptr);
  var
    tx,ty,dx,dy,sh,sw : integer;
    s : string;
    b : boolean;
begin
  showmode ('place | abort');
  s := actcoord^.name^.name;
  b := round2xy (actcoord^.uv,tx,ty);
  dx := 0;
  dy := 0;
  getdrawtext (actcoord^.coordtype,s,actcoord^.name^.typec,dx,dy);
  setcolor (bgc);
  xhidecursor;
  drawmark (tx,ty,actcoord^.name^.filetype);
  if displaytext and (s <> '') then
  begin
    moveto (tx+dx,ty+dy);
    outtext (s);
    sw := textwidth (s);
    sh := th;
  end;
  xshowcursor;
  setcolor (fgc);
  redraw := true;
  startfine;
  repeat
    tx := round(cx);
    ty := round(cy);
    newactpos;
    if displaytext and (s <> '') then
    begin
      xhidecursor;
      revbox (tx+dx,ty+dy,tx+dx+sw,ty+dy+sh);
      xshowcursor;
    end;
    delay (mousedel);
    updatefine (true);
    if displaytext and (s <> '') then
    begin
      xhidecursor;
      revbox (tx+dx,ty+dy,tx+dx+sw,ty+dy+sh);
      xshowcursor;
    end;
  until anyclick;
  newactpos;
  endfine;
  if clickedlkey then
  begin
    actcoord^.ll := ll;
    ll2xy (actcoord^.ll,actcoord^.xy);
    xy2uv (actcoord^.xy,actcoord^.uv);
    filemod[actcoord^.name^.filetype] := true;
    redraw := true;
  end;
end;

procedure editedit (actcoord : coordptr; modify : boolean);
  var
    s : string;
    b : boolean;
begin
  openwindow;
  if modify then
    s := 'edit '
  else
    s := '';
  s := s+filetypename[actcoord^.name^.filetype]+' '+
       coordtypename[actcoord^.coordtype];
  newtext (winxc,winyr[1],true,s);
  b := editcoord (actcoord,modify,true);
  closewindow;
end;

procedure editinsert (actcoord : coordptr);
  var
    lx,ly,lxp,lyp,lxs,lys : integer;
    rubbers,b : boolean;
    tmpcoord : coordptr;
begin
  showmode ('place | abort');
  b := round2xy (actcoord^.uv,lx,ly);
  lxp := lx;
  lyp := ly;
  if actcoord^.coordtype in [cline,cpoly] then
  begin
    rubbers := true;
    tmpcoord := actcoord^.succ;
  end
  else
    if actcoord^.coordtype = cpolyend then
    begin
      rubbers := true;
      tmpcoord := actcoord^.name^.coord;
    end
    else
      rubbers := false;
  if rubbers then
    rubbers := round2xy (tmpcoord^.uv,lxs,lys);
  setcolor (bgc);
  xhidecursor;
  if rubbers then
    line (lx,ly,lxs,lys);
  xshowcursor;
  setcolor (fgc);
  redraw := true;
  startfine;
  repeat
    lx := round(cx);
    ly := round(cy);
    newactpos;
    xhidecursor;
    revline (lxp,lyp,lx,ly);
    if rubbers and ((lxp <> lxs) or (lyp <> lys)) then
      revline (lx,ly,lxs,lys);
    xshowcursor;
    delay (mousedel);
    updatefine (true);
    xhidecursor;
    revline (lxp,lyp,lx,ly);
    if rubbers and ((lxp <> lxs) or (lyp <> lys)) then
      revline (lx,ly,lxs,lys);
    xshowcursor;
  until anyclick;
  newactpos;
  endfine;
  if clickedlkey then
  begin
    new (tmpcoord);
    tmpcoord^.name := actcoord^.name;
    tmpcoord^.ll := ll;
    ll2xy (tmpcoord^.ll,tmpcoord^.xy);
    xy2uv (tmpcoord^.xy,tmpcoord^.uv);
    tmpcoord^.coordtype := actcoord^.coordtype;
    insertcoord (tmpcoord,actcoord);
    if actcoord^.coordtype = clineend then
      actcoord^.coordtype := cline;
    if actcoord^.coordtype = cpolyend then
      actcoord^.coordtype := cpoly;
    filemod[actcoord^.name^.filetype] := true;
    redraw := true;
  end;
end;

procedure editremove (actcoord : coordptr);
  var
    s : string;
begin
  if not (actcoord^.coordtype in [cline,clineend,cpoly,cpolyend]) then
  begin
    errorbeep;
    exit;
  end;
  s := 'remove ';
  case actcoord^.name^.filetype of
    fcha : s := s+'chart ';
    flog : s := s+'log ';
  end;
  s := s+'line point?';
  if not confirm (s,true) then
    exit;
  if actcoord^.coordtype in [clineend,cpolyend] then
  begin
    if actcoord^.name^.coord = actcoord then
    begin
      removeblock (actcoord^.name);
      dispose (actcoord^.name);
    end
    else
      actcoord^.pred^.coordtype := actcoord^.coordtype;
  end
  else
    if actcoord^.name^.coord = actcoord then
      actcoord^.name^.coord := actcoord^.succ;
  removecoord (actcoord);
  dispose (actcoord);
  filemod[actcoord^.name^.filetype] := true;
  redraw := true;
end;

procedure editdelete (actcoord : coordptr);
  var
    nextcoord : coordptr;
    s : string;
begin
  s := 'delete '+filetypename[actcoord^.name^.filetype]+' '+
       coordtypename[actcoord^.coordtype]+'?';
  if not confirm (s,true) then
    exit;
  actcoord := actcoord^.name^.coord;
  while actcoord^.coordtype in [cline,cpoly] do
  begin
    nextcoord := actcoord^.succ;
    removecoord (actcoord);
    dispose (actcoord);
    actcoord := nextcoord;
  end;
  removeblock (actcoord^.name);
  dispose (actcoord^.name);
  removecoord (actcoord);
  dispose (actcoord);
  filemod[actcoord^.name^.filetype] := true;
  redraw := true;
end;


{ gps procedures }

procedure navinitgps;
  var
    s,s1,s2 : string;
    c1,c2 : char;
    ok : boolean;

  function str2list : boolean;
    var
      i,j : integer;
  begin
    str2list := false;
    gpsmsgnum := length(gpsmsgstr);
    if (gpsmsgnum < 1) or (gpsmsgnum > 10) then
      exit;
    for i := 1 to gpsmsgnum do
    begin
      gpsmsglist[i] := 0;
      for j := 1 to 6 do
        if gpsmsgchar[j] <> ' ' then
          if gpsmsgstr[i] = gpsmsgchar[j] then
            gpsmsglist[i] := j;
      if gpsmsglist[i] = 0 then
        exit;
    end;
    str2list := true;
  end;

begin
  if not gpson then
  begin
    openwindow;
    newtext (winxc,winyr[1],true,'init gps');
    s1 := 'gps data format [Nmea0183,Gpspac]: ';
    if gpsnmea then
      c1 := 'N'
    else
      c1 := 'G';
    s2 := 'serial interface (com) [1,2]     : ';
    if gpscom = 1 then
      c2 := '1'
    else
      c2 := '2';
    newtext (winx1+twb,winyr[3],false,s1+c1);
    newtext (winx1+twb,winyr[4],false,s2+c2);
    repeat
      if not editchar (winx1+twb,winyr[3],s1,c1) then
      begin
        closewindow;
        exit;
      end;
      ok := (c1 in ['N','G','n','g']);
      if ok then
        gpsnmea := (c1 in ['N','n'])
      else
        errorbeep;
    until ok;
    repeat
      if not editchar (winx1+twb,winyr[4],s2,c2) then
      begin
        closewindow;
        exit;
      end;
      ok := (c2 in ['1','2']);
      if ok then
      begin
        if c2 = '1' then
          gpscom := 1
        else
          gpscom := 2;
      end
      else
        errorbeep;
    until ok;
    if gpsnmea then
    begin
      newtext (winx1+twb,winyr[7],false,
               '                                   '+
               '(p=GPGLL v=GPVTG c=GPRMC e=PGRME)');
      s := 'decoded gps sentences:             ';
      s1 := gpsmsgstr;
      repeat
        if not editline (winx1+twb,winyr[6],s,s1) then
        begin
          closewindow;
          exit;
        end;
        gpsmsgstr := s1;
        ok := str2list;
        if not ok then
          errorbeep;
      until ok;
      gpsmsgidx := 1;
    end;
    if not confirm ('start gps?',true) then
    begin
      closewindow;
      exit;
    end;
    closewindow;
    if oldsersaved then
      restoreser;
    if gpsnmea then
      initnmea (gpscom)
    else
      initgps (gpscom);
    oldsersaved := true;
    gpsbytesleft := -1;
    gpstime := -1;
    gpson := true;
    gpscog := 0;
    gpssog := 0;
    gpsestimerr := 0;
    gpsmsg := 'gps wait';
    if showgps then
      newgpsdata (false);
    oldusemouse := usemouse;
    xhidecursor;
    usemouse := false;
    xshowcursor;
  end
  else   { gpson }
  begin
    s := 'end gps';
    if gpslogon then
      s := s+' & gps log';
    if not confirm (s+'?',true) then
      exit;
    gpsbytesleft := -1;
    if oldsersaved then
    begin
      restoreser;
      oldsersaved := false;
    end;
    gpslogon := false;
    gpson := false;
    gpstime := -1;
    gpscog := 0;
    gpssog := 0;
    gpsestimerr := 0;
    gpsmsg := 'gps off';
    gpsmarktoggle := true;
    if showgps then
      newgpsdata (false);
    xhidecursor;
    usemouse := oldusemouse;
    xshowcursor;
  end;
end;

procedure navinfogps;
  var
    s,ss : string;
    ll : llt;
    i,j : integer;
begin
  if not gpson then
  begin
    errorbeep;
    exit;
  end;
  openwindow;
  if gpsnmea then
  begin
    newtext (winxc,winyr[1],true,'gps data (nmea0183)');
    j := 4;
    repeat
      repeat
      until keypressed or readserok;
      if readserok then
      begin
        i := readnmea (s);
        { log disabled here }
        ss := '???';
        case i of
          0  : ss := 'timeout                  ';
          -1 : ss := 'error                    ';
          -2 : ss := 'checksum error           ';
          1  : ss := 'sentence without checksum';
          2  : ss := 'sentence with checksum   ';
        end;
        newtext (winx1+twb,winyr[3],false,ss);
        if i <> 0 then
        begin
          newtext (winx1+twb,winyr[j],false,s+'     ');
          inc (j);
          if j = 10 then
            j := 4;
        end;
      end;
    until keypressed;
  end
  else
  begin
    newtext (winxc,winyr[1],true,'gps data (gpspac)');
    repeat
      repeat
      until keypressed or readserok;
      if readserok then
      begin
        gpsbytesleft := readgps;
        gpschecklog (false);
        str (gpsbytesleft:3,s);
        str (gpstimeout:5,ss);
        newtext (winx1+twb,winyr[3],false,'bytes left: '+s+
                                          '  timeout: '+ss+'     ');
        if gpsbytesleft = 0 then
        begin
          getgpsdata;
          newtext (winx1+twb,winyr[4],false,'time:   '+getgpstimestr+'     ');
          ll.lat := gpslat;
          ll.lon := gpslon;
          newtext (winx1+twb,winyr[5],false,'pos:    '+ll2str(ll)+'     ');
          str (gpsvele:10:6,s);
          str (gpsveln:10:6,ss);
          newtext (winx1+twb,winyr[6],false,'vel e:  '+s+' m/s'+
                                            '   vel n:  '+ss+' m/s     ');
          str (gpsalt:10:3,s);
          str (gpsvelup:10:6,ss);
          newtext (winx1+twb,winyr[7],false,'alt:    '+s+' m  '+
                                            '   vel up: '+ss+' m/s     ');
          str (gpsehpe:10:0,s);
          str (gpsevpe:10:0,ss);
          newtext (winx1+twb,winyr[8],false,'ehpe:   '+s+' m  '+
                                            '   evpe:   '+ss+' m     ');
          str (gpsete*1e6:10:3,s);
          str (gpsehv:10:2,ss);
          newtext (winx1+twb,winyr[9],false,'ete:    '+s+' us '+
                                            '   ehv:    '+ss+' m/s     ');
        end;
      end;
    until keypressed;
  end;
  repeat
    delay (mousedel);
    updatefine (false);
  until anyclick;
  closewindow;
  if gpslogon then
    redraw := true;
end;

procedure navwptsgps;
  var
    i : integer;
    s : string;
    b : boolean;
    numwpts : integer;
    name,name1 : nmeawpt;
    lat,lon : real;
    coord : coordptr;
    block : blockptr;
    gotall : boolean;
begin
  if (not gpson) or (not gpsnmea) then
  begin
    errorbeep;
    exit;
  end;
  openwindow;
  newtext (winxc,winyr[1],true,'receive nmea waypoints');
  numwpts := 0;
  gotall := false;
  repeat
    repeat
    until keypressed or readserok;
    if readserok then
    begin
      i := readnmea (s);
      { log disabled here }
      if i > 0 then
      begin
        i := getnmeatype(s);
        if i = 3 then
          if getnmeawpl (s,lat,lon,name) then
          begin
            if numwpts = 0 then
              name1 := name;
            if (numwpts > 0) and (name = name1) then
              gotall := true
            else
            begin
              new (coord);
              new (block);
              coord^.name := block;
              coord^.ll.lat := lat;
              coord^.ll.lon := lon;
              ll2xy (coord^.ll,coord^.xy);
              xy2uv (coord^.xy,coord^.uv);
              coord^.coordtype := cmark;
              block^.coord := coord;
              block^.name := name;
              block^.filetype := fwpt;
              block^.scalelimit := 0;   { defscalelimit; }
              block^.typec := ' ';
              defscalelimit := block^.scalelimit;
              insertcoord (coord,coordlast);
              insertblock (block,blocklast);
              filemod[fwpt] := true;
              redraw := true;
              inc (numwpts);
              str (numwpts,s);
              newtext (winx1+twb,winyr[3],false,'waypoint '+s+': '
                       +name+'        ');
            end;
          end;
        end;
    end;
  until gotall or keypressed;
  str (numwpts,s);
  if gotall then
    b := confirm ('received '+s+' waypoints',false);
  closewindow;
end;

procedure navloggps;
  var
    s,ss : string;
    tmpblock : blockptr;
    b : boolean;
    i,er : integer;
begin
  if not gpson then
  begin
    errorbeep;
    exit;
  end;
  if not gpslogon then
  begin
    if ((not gpsnmea) and (gpsbytesleft <> 0)) or
       (gpsnmea and (gpstime < 0)) then
    begin
      errorbeep;
      b := confirm ('Error: no gps data',false);
      exit;
    end;
    openwindow;
    newtext (winxc,winyr[1],true,'gps log');
    s := 'log delta t [s]: ';
    str (gpslogtime,ss);
    repeat
      if not editline (winx1+twb,winyr[3],s,ss) then
      begin
        closewindow;
        exit;
      end;
      val (ss,i,er);
      if er = 0 then
      begin
        if i >= 1 then
          gpslogtime := i
        else
          er := 1;
      end;
      if er <> 0 then
        errorbeep;
    until er = 0;
    if not confirm ('start gps log?',true) then
    begin
      closewindow;
      exit;
    end;
    closewindow;
    gpslogon := true;
    gpsnextlogtime := gpstime + gpslogtime;
    if gpsnextlogtime > daysec then
      gpsnextlogtime := gpsnextlogtime - daysec;
    s := 'log start: '+getgpstimestr;
    if gpsnmea then
      s := s+' (local) '
    else
      s := s+' (UTC) ';
    str (gpslogtime,ss);
    s := s+'  dt: '+ss+'s';
    new (tmpblock);
    new (gpsloglastcoord);
    insertblock (tmpblock,blocklast);
    insertcoord (gpsloglastcoord,coordlast);
    filemod[flog] := true;
    tmpblock^.name := s;
    tmpblock^.coord := gpsloglastcoord;
    tmpblock^.typec := ' ';
    tmpblock^.scalelimit := defscalelimit;
    tmpblock^.filetype := flog;
    gpsloglastcoord^.ll := gpsll;
    ll2xy (gpsloglastcoord^.ll,gpsloglastcoord^.xy);
    xy2uv (gpsloglastcoord^.xy,gpsloglastcoord^.uv);
    gpsloglastcoord^.name := tmpblock;
    gpsloglastcoord^.coordtype := clineend;
    gpsmsg := 'log on';
    if showgps then
      newgpsdata (false);
  end
  else
  begin
    if confirm ('end gps log?',true) then
    begin
      gpslogon := false;
      gpsmsg := 'log off';
      redraw := true;
    end;
  end;
end;


{ main menus }

procedure domenumenu;
  var
    c : char;
    s : string;
begin
  drawmenu (menumenu);
  c := getmenufp (menumenu);
  case c of
    'f' : actmenu := c;
    'd' : actmenu := c;
    'n' : actmenu := c;
    'e' : actmenu := c;
    'h' : showhelp;
    'r' : recalc := true;
    'q' : begin
            s := '';
            if filemod[fcha] then
              s := s+','+filetypename[fcha];
            if filemod[flog] then
              s := s+','+filetypename[flog];
            if filemod[fmem] then
              s := s+','+filetypename[fmem];
            if filemod[fwpt] then
              s := s+','+filetypename[fwpt];
            if s <> '' then
            begin
              delete (s,1,1);
              doquit := confirm (s+' data not saved! quit?',true);
            end
            else
              doquit := true;
            if doquit then
              doquit := confirm ('quit '+title+'?',true);
          end;
  end;
  freemenu (menumenu);
end;

procedure commonkeys (c : char);
begin
  case c of
    esc : drawbigcross;
    'O' : zoomfact (2);
    'I' : zoomfact (0.5);
    'M' : transmove (ll);
    'L' : begin
            bmpdisp := bmpdisplast;
            restoreab;
            setchartsize;
            recalc := true;
          end;
    'G' : if gpson then
            errorbeep
          else
          begin
            gpsll := ll;
            gpssog := 0;
            gpscog := 0;
            gpsestimerr := 0;
          end;
    'C' : transmove (gpsll);
  end;
end;

procedure filemenu;
  var
    c1,c2 : char;
    b : boolean;
begin
  top3x := menufile.x2+1;
  initactpos;
  showmode ('  | +');
  if showgps then
    initgpsdata;
  drawmenu (menufile);
  c1 := getmenufp (menufile);
  commonkeys (c1);
  case c1 of
    'm' : domenumenu;
    'l' : begin
            drawmenu (menufload);
            c2 := getmenufp (menufload);
            if c2 in ['c','o','l','m','w'] then
            begin
              openwindow;
              newtext (winxc,winyr[1],true,'load/merge data');
              case c2 of
                'c' : b := loadsave (fcha,true);
                'o' : if loadsave (fovl,true) then
                        displayovl := true;
                'l' : b := loadsave (flog,true);
                'm' : b := loadsave (fmem,true);
                'w' : b := loadsave (fwpt,true);
              end;
              closewindow;
            end;
            freemenu (menufload);
          end;
    's' : begin
            drawmenu (menufsave);
            c2 := getmenufp (menufsave);
            if c2 in ['c','l','m','w'] then
            begin
              openwindow;
              newtext (winxc,winyr[1],true,'save data');
              initll2str (fllformat,flllength);
              case c2 of
                'c' : b := loadsave (fcha,false);
                'l' : b := loadsave (flog,false);
                'm' : b := loadsave (fmem,false);
                'w' : b := loadsave (fwpt,false);
              end;
              initll2str (llformat,lllength);
              closewindow;
            end;
            freemenu (menufsave);
          end;
    'c' : begin
            drawmenu (menufclear);
            c2 := getmenufp (menufclear);
            case c2 of
              'c' : if confirm('clear chart data?',true) then
                      cleard (fcha);
              'o' : if confirm('clear overlay data?',true) then
                      cleard (fovl);
              'l' : if confirm('clear log data?',true) then
                      cleard (flog);
              'm' : if confirm('clear memo data?',true) then
                      cleard (fmem);
              'w' : if confirm('clear waypoint data?',true) then
                      cleard (fwpt);
              'a' : if confirm('clear all data (cha,ovl,log,mem,wpt)?',true) then
                    begin
                      cleard (fcha);
                      cleard (fovl);
                      cleard (flog);
                      cleard (fmem);
                      cleard (fwpt);
                    end;
            end;
            freemenu (menufclear);
          end;
    'i' : mainshowinfo;
    'f' : fileformat;
    'b' : begin
            drawmenu (menufbmp);
            c2 := getmenufp (menufbmp);
            case c2 of
              'a' : filebmpalign;
              'l' : begin
                      filebmpload;
                      if bmpnameload <> '' then
                        filebmpreload;
                    end;
              'r' : begin
                      bmpcenterll := currentchartcenter;
                      filebmpreload;
                    end;
              'c' : begin
                      bmpdisp := false;
                      redraw := true;
                    end;
              's' : filebmpsavescreen;
            end;
            freemenu (menufbmp);
          end;
  end;
end;

procedure displaymenu;
  var
    c1,c2 : char;
    b : boolean;
begin
  top3x := menudisp.x2+1;
  initactpos;
  showmode ('move | +');
  if showgps then
    initgpsdata;
  drawmenu (menudisp);
  c1 := getmenufp (menudisp);
  commonkeys (c1);
  case c1 of
    ' ' : transmove (ll);
    'm' : domenumenu;
    's' : begin
            drawmenu (menudscale);
            c2 := getmenufp (menudscale);
            case c2 of
              'f' : inittransdigit;
              'a' : inittransalign;
              'c' : inittranschart;
              'm' : inittransmercator;
              'l' : begin
                      bmpdisp := bmpdisplast;
                      restoreab;
                      setchartsize;
                      recalc := true;
                    end;
            end;
            freemenu (menudscale);
          end;
    'i' : zoomfact (0.5);
    'o' : zoomfact (2);
    'z' : zoommouse;
    'c' : begin
            drawmenu (menudconfig);
            c2 := getmenufp (menudconfig);
            case c2 of
              'f' : displlformat;
              'd' : dispgps;
              'g' : dispdrawgrid;
              'i' : dispdisplayed;
              'r' : begin
                      reverse := not reverse;
                      initcolor;
                      initfinecolor;
                      redraw := true;
                    end;
              'm' : if usemouse then
                      dispmousecursor
                    else
                      b := confirm ('no mouse installed',false);
            end;
            freemenu (menudconfig);
          end;
  end;
end;

procedure navigatemenu;
  var
    c1 : char;
begin
  top3x := menunav.x2+1;
  initactpos;
  showmode ('  | +');
  if showgps then
    initgpsdata;
  drawmenu (menunav);
  c1 := getmenufp (menunav);
  commonkeys (c1);
  case c1 of
    'm' : domenumenu;
    'g' : navinitgps;
    'i' : navinfogps;
    'w' : navwptsgps;
    'l' : navloggps;
    'c' : transmove (gpsll);
    'd' : dispgps;
  end;
end;

procedure editmenu;
  var
    c1,c2 : char;
    isovl : boolean;
begin
  top3x := menuedit.x2+1;
  initactpos;
  showmode (editmodename[editmode]);
  if showgps then
    initgpsdata;
  drawmenu (menuedit);
  c1 := getmenufp (menuedit);
  commonkeys (c1);
  case c1 of
    ' ' : begin
            if editmode in [maddchaline,maddchapoly,maddchatext,
                            maddchamark,maddlog,maddmem,maddwpt] then
            begin
              case editmode of
                maddchaline : addlinepoly (fcha,cline);
                maddchapoly : addlinepoly (fcha,cpoly);
                maddchatext : addmarktext (fcha,ctext);
                maddchamark : addmarktext (fcha,cmark);
                maddlog     : addlinepoly (flog,cline);
                maddmem     : addmarktext (fmem,cmark);
                maddwpt     : addmarktext (fwpt,cmark);
              end;
            end
            else
            begin
              actcoord := findcoord (cx,cy);
              if actcoord = nil then
                errorbeep
              else
              begin
                isovl := (actcoord^.name^.filetype = fovl);
                case editmode of
                  mmove   : if isovl then
                              errorbeep
                            else
                              if actcoord^.coordtype in [ctext,cmark] then
                                editmovemarktext (actcoord)
                              else
                                editmovelinepoly (actcoord);
                  medit   : editedit (actcoord,not isovl);
                  minsert : if isovl or not (actcoord^.coordtype in
                               [cline,clineend,cpoly,cpolyend]) then
                              errorbeep
                            else
                              editinsert (actcoord);
                  mremove : if isovl or not (actcoord^.coordtype in
                               [cline,clineend,cpoly,cpolyend]) then
                              errorbeep
                            else
                              editremove (actcoord);
                  mdelete : if isovl then
                              errorbeep
                            else
                              editdelete (actcoord);
                end;
              end;
            end;
          end;
    'm' : domenumenu;
    'a' : begin
            drawmenu (menueadd);
            c2 := getmenufp (menueadd);
            case c2 of
              'l' : seteditmode (maddchaline);
              'p' : seteditmode (maddchapoly);
              't' : seteditmode (maddchatext);
              'm' : seteditmode (maddchamark);
              'o' : seteditmode (maddlog);
              'e' : seteditmode (maddmem);
              'w' : seteditmode (maddwpt);
            end;
            freemenu (menueadd);
          end;
    'e' : begin
            drawmenu (menueedit);
            c2 := getmenufp (menueedit);
            case c2 of
              'e' : seteditmode (medit);
              'i' : seteditmode (minsert);
              'r' : seteditmode (mremove);
              'd' : seteditmode (mdelete);
            end;
            freemenu (menueedit);
          end;
    'v' : seteditmode (mmove);
  end;
end;


{ main program }

begin   { main }
  if not initdev then
    exit;
  init;
  seteditmode (maddchaline);
  recalc := false;
  redraw := false;
  doquit := false;
  titlescreen;
  while not doquit do
  begin
    if recalc then
    begin
      transcoord;
      redraw := true;
    end;
    if redraw then
      drawchart;
    recalc := false;
    redraw := false;
    case actmenu of
      'f' : filemenu;
      'd' : displaymenu;
      'n' : navigatemenu;
      'e' : editmenu;
    else
      doquit := true;   { error }
    end;
  end;
  if oldsersaved then
    restoreser;
  restorecrtmode;
  closegraph;
end.   { main }
