unit grafbmp;

{ grafbmp.pas }
{ .bmp bitmap file read/write }

{ (c) Heiko Purnhagen   23-jul-96 }
{ HP 960725 }

interface

uses
  graph;

type
  bytefile = file of byte;

function readbmphdr (filename : string;
                     var wid,hig,col : integer; var c00 : boolean) : string;
{ ok->'' error->'<error msg>' }

function readbmp (filename : string; xoff,yoff : integer;
                  x0,y0,xs,ys,zoom,thr,bgc,fgc : integer) : string;
{ ok->'' error->'<error msg>' }
{ xoff,yoff=bmp offset (lower left corner) }
{ x0,y0=lower left corner on screen }
{ xs,ys=size on screen }
{ zoom=-4->*1/16 zoom=0->*1 zoom=2->*4   thr=1->norm thr=-1->inv }

function writebmp (filename : string; x0,y0,xs,ys : integer) : string;
{ ok->'' error->'<error msg>' }
{ x0,y0=lower left corner on screen }
{ xs,ys=size on screen }


implementation

function readbmpword (var f : bytefile;
                      numword : integer; var w : word) : boolean;
{ ok->true error->false }
  var
    i : integer;
    b : byte;
begin
  readbmpword := false;
  w := 0;
  if numword >= 1 then
  begin
    {$i-}
    read (f,b);
    {$i+}
    if ioresult <> 0 then
      exit;
    w := b;
    {$i-}
    read (f,b);
    {$i+}
    if ioresult <> 0 then
      exit;
    w := w+(b shl 8);
  end;
  for i := 3 to numword*2 do
  begin
    {$i-}
    read (f,b);
    {$i+}
    if ioresult <> 0 then
      exit;
  end;
  readbmpword := true;
end;

function writebmpword (var f : bytefile;
                       numword : integer; w : word) : boolean;
{ ok->true error->false }
  var
    i : integer;
    b : byte;
begin
  writebmpword := false;
  if numword >= 1 then
  begin
    b := lo(w);
    {$i-}
    write (f,b);
    {$i+}
    if ioresult <> 0 then
      exit;
    b := hi(w);
    {$i-}
    write (f,b);
    {$i+}
    if ioresult <> 0 then
      exit;
  end;
  b := 0;
  for i := 3 to numword*2 do
  begin
    {$i-}
    write (f,b);
    {$i+}
    if ioresult <> 0 then
      exit;
  end;
  writebmpword := true;
end;

function readbmphdr (filename : string;
                     var wid,hig,col : integer; var c00 : boolean) : string;
{ ok->'' error->'<error msg>' }
  var
    f : bytefile;
    w : word;
    i : integer;
begin
  assign (f,filename);
  {$i-}
  reset (f);
  {$i+}
  if ioresult <> 0 then
  begin
    readbmphdr := 'can''t open file';
    exit;
  end;
  if not readbmpword (f,1,w) then
  begin
    readbmphdr := 'file too short';
    close (f);
    exit;
  end;
  if w <> $4d42 then
  begin
    readbmphdr := 'no bmp header';
    close (f);
    exit;
  end;
  if not readbmpword (f,8,w) then
  begin
    readbmphdr := 'file too short';
    close (f);
    exit;
  end;
  if not readbmpword (f,1,w) then
  begin
    readbmphdr := 'file too short';
    close (f);
    exit;
  end;
  wid := w;
  if not readbmpword (f,1,w) then
  begin
    readbmphdr := 'file too short';
    close (f);
    exit;
  end;
  if not readbmpword (f,1,w) then
  begin
    readbmphdr := 'file too short';
    close (f);
    exit;
  end;
  hig := w;
  if not readbmpword (f,1,w) then
  begin
    readbmphdr := 'file too short';
    close (f);
    exit;
  end;
  if not readbmpword (f,1,w) then
  begin
    readbmphdr := 'file too short';
    close (f);
    exit;
  end;
  if w <> $0001 then
  begin
    readbmphdr := 'more than 1 layer';
    close (f);
    exit;
  end;
  if not readbmpword (f,1,w) then
  begin
    readbmphdr := 'file too short';
    close (f);
    exit;
  end;
  col := 1 shl w;
  if col > 256 then
  begin
    readbmphdr := 'more than 256 colors';
    close (f);
    exit;
  end;
  if not readbmpword (f,12,w) then
  begin
    readbmphdr := 'file too short';
    close (f);
    exit;
  end;
  { rgb palette }
  if not readbmpword (f,1,w) then
  begin
    readbmphdr := 'file too short';
    close (f);
    exit;
  end;
  c00 := (w = 0);
  if not readbmpword (f,1,w) then
  begin
    readbmphdr := 'file too short';
    close (f);
    exit;
  end;
  c00 := c00 and (w = 0);
  close (f);
  readbmphdr := '';
end;

function readbmp (filename : string; xoff,yoff : integer;
                  x0,y0,xs,ys,zoom,thr,bgc,fgc : integer) : string;
{ ok->'' error->'<error msg>' }
{ xoff,yoff=bmp offset (lower left corner) }
{ x0,y0=lower left corner on screen }
{ xs,ys=size on screen }
{ zoom=-4->*1/16 zoom=0->*1 zoom=2->*4   thr=1->norm thr=-1->inv }
  var
    f : bytefile;
    w : word;
    b : byte;
    i : integer;
    wid,hig,col,colbit,colmask,colshift : integer;
    byteline,pelbyte,bytedisp : integer;
    bytepre,bytepost : integer;
    r,c,cc,cd,x,y,xd,xi,yi,pelsize : integer;
    pelcol,pelcol0,pelcol1 : integer;
begin
  assign (f,filename);
  {$i-}
  reset (f);
  {$i+}
  if ioresult <> 0 then
  begin
    readbmp := 'can''t open file';
    exit;
  end;
  if not readbmpword (f,1,w) then
  begin
    readbmp := 'file too short';
    close (f);
    exit;
  end;
  if w <> $4d42 then
  begin
    readbmp := 'no bmp header';
    close (f);
    exit;
  end;
  if not readbmpword (f,8,w) then
  begin
    readbmp := 'file too short';
    close (f);
    exit;
  end;
  if not readbmpword (f,1,w) then
  begin
    readbmp := 'file too short';
    close (f);
    exit;
  end;
  wid := w;
  if not readbmpword (f,1,w) then
  begin
    readbmp := 'file too short';
    close (f);
    exit;
  end;
  if not readbmpword (f,1,w) then
  begin
    readbmp := 'file too short';
    close (f);
    exit;
  end;
  hig := w;
  if not readbmpword (f,1,w) then
  begin
    readbmp := 'file too short';
    close (f);
    exit;
  end;
  if not readbmpword (f,1,w) then
  begin
    readbmp := 'file too short';
    close (f);
    exit;
  end;
  if w <> $0001 then
  begin
    readbmp := 'more than 1 layer';
    close (f);
    exit;
  end;
  if not readbmpword (f,1,w) then
  begin
    readbmp := 'file too short';
    close (f);
    exit;
  end;
  colbit := w;
  col := 1 shl w;
  colmask := col-1;
  if col > 256 then
  begin
    readbmp := 'more than 256 colors';
    close (f);
    exit;
  end;
  if not readbmpword (f,12,w) then
  begin
    readbmp := 'file too short';
    close (f);
    exit;
  end;
  { rgb palette }
  if not readbmpword (f,col*2,w) then
  begin
    readbmp := 'file too short';
    close (f);
    exit;
  end;
  byteline := ((wid*colbit+31) div 32)*4;
  pelcol0 := bgc;
  pelcol1 := fgc;
  if thr < 0 then
  begin
    thr := -thr;
    pelcol0 := fgc;
    pelcol1 := bgc;
  end;
  colmask := colmask shl (8-colbit);
  colshift := colbit;
  cd := 1;
  xd := 1;
  bytedisp := 1;
  if zoom > 2 then
    zoom := 2;
  if zoom < 0 then
  begin
    cd := 1 shl -zoom;
    colshift := colbit shl -zoom;
    bytedisp := (colshift+7) div 8
  end;
  if zoom > 0 then
    xd := 1 shl zoom;
  pelbyte := 8 div colbit;
  bytepre := xoff div pelbyte;
  bytepost := byteline-1-
              ((((xs shl (2-zoom)) shr 2)+xoff) div pelbyte);
  if bytepre < 0 then
    bytepre := 0;
  if bytepost < 0 then
    bytepost := 0;
  thr := thr shl (8-colbit);
  if zoom > 0 then
    pelsize := (1 shl zoom)-1
  else
    pelsize := 0;
  for r := 0-yoff to hig-1-yoff do
  begin
    if r < 0 then
      y := y0+1
    else
      y := y0-(((r) shl 2) shr (2-zoom));
    if (y >= y0) or (y-pelsize <= y0-ys) or
       ((zoom < 0) and (r mod (1 shl -zoom) <> 0)) then
    begin
      {$i-}
      seek (f,filepos(f)+byteline);
      {$i+}
      if ioresult <> 0 then
      begin
        readbmp := 'file too short';
        close (f);
        exit;
      end;
    end
    else
    begin
      if bytepre > 0 then
      begin
        {$i-}
        seek (f,filepos(f)+bytepre);
        {$i+}
        if ioresult <> 0 then
        begin
          readbmp := 'file too short';
          close (f);
          exit;
        end;
      end;
      c := bytepre*pelbyte-xoff;
      for i := 1 to (byteline-bytepre-bytepost) div bytedisp do
      begin
        {$i-}
        read (f,b);
        {$i+}
        if ioresult <> 0 then
        begin
          readbmp := 'file too short';
          close (f);
          exit;
        end;
        x := x0+((c shl 2) shr (2-zoom));
        cc := c+pelbyte;
        while c < cc do
        begin
          if ((b and colmask) < thr) then
            pelcol := pelcol0
          else
            pelcol := pelcol1;
          if pelsize = 0 then
            putpixel (x,y,pelcol)
          else
            for xi := 0 to pelsize do
              for yi := 0 to pelsize do
                putpixel (x+xi,y-yi,pelcol);
          b := b shl colshift;
          c := c+cd;
          x := x+xd;
        end;
        if bytedisp > 1 then
        begin
          {$i-}
          seek (f,filepos(f)+bytedisp-1);
          {$i+}
          if ioresult <> 0 then
          begin
            readbmp := 'file too short';
            close (f);
            exit;
          end;
        end;
      end;
      if bytepost > 0 then
      begin
        {$i-}
        seek (f,filepos(f)+bytepost);
        {$i+}
        if ioresult <> 0 then
        begin
          readbmp := 'file too short';
          close (f);
          exit;
        end;
      end;
    end;
  end;
  close (f);
  readbmp := '';
end;

function writebmp (filename : string; x0,y0,xs,ys : integer) : string;
{ ok->'' error->'<error msg>' }
{ x0,y0=lower left corner on screen }
{ xs,ys=size on screen }
  var
    f : bytefile;
    w : word;
    b : byte;
    i,j : integer;
    byteline : integer;
    x,y : integer;
    pel : integer;
begin
  assign (f,filename);
  {$i-}
  rewrite (f);
  {$i+}
  if ioresult <> 0 then
  begin
    writebmp := 'can''t open file';
    exit;
  end;
  if not writebmpword (f,1,$4d42) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,1,$007e) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,3,0) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,1,$003e) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,1,0) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,1,$0028) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,1,0) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,1,xs) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,1,0) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,1,ys) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,1,0) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,1,1) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,1,1) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,12,0) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,1,$0000) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,1,$0000) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,1,$ffff) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  if not writebmpword (f,1,$00ff) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;

  byteline := ((xs+31) div 32)*4;
  for y := 0 to ys-1 do
  begin
    b := 0;
    j := 0;
    for x := 0 to xs-1 do
    begin
      b := b shl 1;
      if getpixel(x0+x,y0-y) > 0 then
        b := b or 1;
      if x mod 8 = 7 then
      begin
        {$i-}
        write (f,b);
        {$i+}
        if ioresult <> 0 then
        begin
          writebmp := 'error writing file';
          close (f);
          exit;
        end;
        inc (j);
        b := 0;
      end;
    end;
    if xs mod 8 <> 0 then
    begin
      b := b shl (8-(xs mod 8));
      {$i-}
      write (f,b);
      {$i+}
      if ioresult <> 0 then
      begin
        writebmp := 'error writing file';
        close (f);
        exit;
      end;
      inc (j);
    end;
    b := 0;
    while j < byteline do
    begin
      {$i-}
      write (f,b);
      {$i+}
      if ioresult <> 0 then
      begin
        writebmp := 'error writing file';
        close (f);
        exit;
      end;
      inc(j);
    end;
  end;
  close (f);
  writebmp := '';
end;

begin
end.

{ end unit }
